(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Ass;;
open Partition;;

(* Interference graph *)

type node_info = {
    temp : temp;
    mutable color : temp option;
    mutable adj : node list;
    mutable moves : move list;
    mutable degree : int;
    mutable alias : node option;
    mutable occurs : int;
  }
and move_info = {move: Ass.instr; mutable left : node; mutable right : node}
and node = node_info elem
and move = move_info elem

(* 
   precolored = obvious 
   initial = during initialization
   simplifiable = low density, no moves
   complex = I might be able to coalesce
   congested = I can simplify to a slow register
   dense = I cannot simplify
   fusioned = fusioned with another register
   stacked = has been simplified
   colored = what any tempory aspires to
*)

(*
   fusioned = dealt with
   constrained = registers interfere
   frozen = given up on
   safe = ready to be fusioned
   slowsafe = fusionable and maintains slow colorability
   unsafe = fusionable if ends were less constrained
   initialm = during initialization
*)

let temp_name t = Codegen.namer t;;
let node_name n = temp_name (info n).temp;;

let print_interference p =
  prerr_string "\nRegister interference*************************************************\n";
  Array.iteri 
    (fun i np ->
      Printf.fprintf stderr "  %s:\n" 
	[|"precolored"; "initial"; "lowdensity"; "meddensity"; "highdensity"; 
	  "fusioned"; "stacked"; "colored"; "readoff"|].(i);
      List.iter
	(function n ->
	  let t = (info n).temp in
	  if t = Codegen.regRA || (info n).adj <> [] then begin
	    Printf.fprintf stderr "    %s %i %s with : " (node_name n) (info n).degree
	      (match (info n).color with 
		None -> ""
	      |	Some t -> Printf.sprintf "colored(%s)" (temp_name t));
	    List.iter 
	      (function n -> Printf.fprintf stderr "%s " (node_name n)) 
	      (info n).adj;
	    Printf.fprintf stderr " %i moves." (List.length (info n).moves);
	    prerr_newline ();
	  end) (list np)) p;
  prerr_string "End of register interference.*****************************************\n\n";;

let print_moves p =
  prerr_string "\nRegister attraction*************************************************\n";
  Array.iteri 
    (fun i p -> 
      Printf.fprintf stderr "  %s:\n" 
	[|"fusionedm"; "constrained"; "frozen"; "safe"; "slowsafe"; "unsafe"; "initialm"|].(i);
      List.iter (function { info = m } as n -> 
	Printf.fprintf stderr "    %s <-> %s\n" (node_name m.left) (node_name m.right) 
	    ) (list p)) p;
  prerr_string "End of register attraction.*****************************************\n\n";;

let alloc fastcolors slowcolors precol code flow_graph =
  flush stdout;
  if Args.verbose > 0 then
    prerr_string "\nStarting register allocation\n";
  (* Hashtables and partitions *)

  let temp2nodes = Hashtbl.create 13 in
  let nodepartition = make 9 in
  let precolored = nodepartition.(0) in
  let initial = nodepartition.(1) in 
  let lowdensity = nodepartition.(2) in 
  let meddensity = nodepartition.(3) in 
  let highdensity = nodepartition.(4) in 
  let fusioned = nodepartition.(5) in 
  let stacked = nodepartition.(6) in 
  let colored = nodepartition.(7) in 
  let readoff = nodepartition.(8) in

  let movepartition = make 7 in
  let fusionedm = movepartition.(0) in 
  let constrained = movepartition.(1) in 
  let frozen = movepartition.(2) in 
  let safe = movepartition.(3) in 
  let slowsafe = movepartition.(4) in 
  let unsafe = movepartition.(5) in 
  let initialm = movepartition.(6) in

  let numfast = List.length fastcolors in
  let numslow = List.length slowcolors in

  let register_picklist = fastcolors @ slowcolors in
  
  let findnode t = Hashtbl.find temp2nodes t in

  let interference () =
    let new_node n = create initial n in
    let node b t = (* Finds a node, creates if necessary (precolored if b) *)
      try Hashtbl.find temp2nodes t
      with Not_found ->
	let i =
	  (*Printf.eprintf "Adding node %s\n" (Codegen.namer t);*)
          new_node 
	    { temp = t; color = if b then Some t else None; adj = []; 
	      degree = 0; alias = None; occurs = 0; moves = [] } 
	in
	Hashtbl.add temp2nodes t i; i in
    List.iter (fun t -> let _ = node true t in ()) precol; (* Add precoloured temporaries *)
    let node = node false in (* Now node is always non precolored *)
    let new_edge (u : node) (v : node) = (* Adds an interference edge *)
      if u = v || List.mem u (info v).adj then ()
      else begin
	(info u).adj <- v ::(info u).adj; 
	(info v).adj <- u :: (info v).adj; 
	(*Printf.eprintf "    added interference between %s and %s\n" (node_name u) (node_name v)*)
      end
    in
    let interference m = (* adds the interferences for an instruction *)
      (* Ass.print stderr Codegen.namer m.instr; prerr_newline (); *)
      let edge r1 r2 = new_edge (node r1) (node r2) in
      let rec add_gen_interf = function
	  h::t -> List.iter (edge h) t; add_gen_interf t
	| [] -> ()
      in
      if not m.dead then
	(match m.instr with
	| Move (s, d) as move ->
            let is = node s and id = node d in
	    let is_fusionable n = 
	      List.mem (info n).temp fastcolors ||
	      List.mem (info n).temp fastcolors
	    in
	    if is_fusionable is && is_fusionable id then
              ignore (create initialm {move = move; left = is; right = id});
	    let equiv_temps = s::d::Codegen.move_parasite_dests in
            List.iter
              (fun def -> 
		List.iter 
		  (fun thru -> 
		    if not (List.mem def equiv_temps) || not (List.mem thru equiv_temps) 
		    then edge def thru) 
		  m.live_out)
              m.def
	      
          (*let edge r = let ir = node r in if is <> ir then new_edge id ir in*)
          (*List.iter edge m.live_out;*)
	| _ ->
            List.iter
              (fun def -> List.iter (edge def) m.live_out)
              m.def) 
    in
    List.iter (fun m ->
      List.iter (fun n -> ignore (node n)) m.def; (* Make sure that all temps have nodes *)
      List.iter (fun n -> ignore (node n)) m.use; (* Make sure that all temps have nodes *)
      ) flow_graph;
    List.iter interference flow_graph in
  
  let add_occurence = function i -> 
    let incr_occurs t = 
      (try 
	let n = findnode t in (info n).occurs <- (info n).occurs + 1
      with Not_found -> raise (Failure ("No node for temporary "^(temp_name t)))); (* Not one of the allocation registers *) 
    in
    List.iter incr_occurs (Codegen.usedby i);
    List.iter incr_occurs (Codegen.defby i)
  in

  let node_order = 
    (fun a b -> (Gen.temp_int (info a).temp) - (Gen.temp_int (info b).temp) ) 
  in

  let rec sorted_list_inclusion l1 l2 = match l1, l2 with
    h1::t1, h2::t2 when h1 = h2 -> sorted_list_inclusion t1 t2
  | h1::_ as l, h2::t when node_order h1 h2 > 0 -> sorted_list_inclusion l t
  | [], _ -> true
  | _ -> false
  in

  let rec union_sorted l1 l2 = match l1, l2 with
    h1::t1, (h2::_ as t2) when h1 = h2 -> union_sorted t1 t2
  | h1::_ as l, h2::t when node_order h1 h2 > 0 -> h2::(union_sorted l t)
  | h1::t, (h2::_ as l) -> h2::(union_sorted t l)
  | [], l -> l
  | l, [] -> l
  in

  let ins_sorted e l = 
    let rec helper = function
	h::t when h = e -> raise Exit
      |	h::t when node_order h e > 0 -> e::h::t
      |	h::t -> h::(helper t)
      |	[] -> [e]
    in
    try helper l with Exit -> l
  in

  let add_interference n1 n2 =
    if Args.verbose > 0 then 
      Printf.eprintf "added interference between %s and %s\n" (node_name n1) (node_name n2);
    (info n1).adj <- ins_sorted n2 (info n1).adj;
    (info n2).adj <- ins_sorted n1 (info n2).adj;
  in
    
  let sort_node_adj n = 
    (*let rec removedup = 
      function 
	  h1::((h2::_) as t) when h1 = h2 -> (removedup t)
	| h::t -> h::(removedup t)
	| [] -> []
    in*)
    n.adj <- List.sort node_order n.adj;
  in

  let nodeclass n =
    let n = (info n) in
    (if n.color <> None then precolored
    else if n.degree < numfast then lowdensity
    else if n.degree < numslow then meddensity
    else highdensity)
  in

  let reclassify n =
    if (belong n lowdensity) || (belong n meddensity) || (belong n highdensity) then
      move n (nodeclass n)
  in

  let active_node n =
    (not (belong n fusioned )) &&
    (not (belong n stacked))
  in

  let calc_degree adj =
    (* Printf.eprintf "calc_degree on ";
    List.iter (fun n -> prerr_string (node_name n); prerr_char ' ') adj; *)
    let colored, notcolored = 
      List.fold_left 
	(fun (cols, i) n -> 
	  if active_node n then begin
	    match (info n).color with
	      None -> cols, i + 1
	    | Some c -> 
		if (List.mem c fastcolors) || (List.mem c slowcolors) then
		  (ins_sorted (Hashtbl.find temp2nodes c) cols), i
		else cols, i
	  end else cols, i)
	([], 0) adj;
    in
    let rslt = List.length colored + notcolored in
    (*Printf.eprintf " -> %i\n" rslt;*)
    rslt
  in

  let classify_node no = 
    let n = info no in (* We do an excessive estimate here because of the precolored nodes *)
    (* Printf.eprintf "Classifying node %s : conflicts with %i uncolored and " (node_name no) notcolored;
    List.iter (fun n -> prerr_string (node_name n); prerr_char ' ') colored;
    prerr_newline ();
    prerr_string "of ";
    List.iter (fun n -> prerr_string (node_name n); prerr_char ' ') n.adj;
    prerr_newline ();*)
    n.degree <- calc_degree n.adj;
    move no (nodeclass no)
  in

  let classify_unsorted_node no = 
    sort_node_adj (info no);
    classify_node no
  in

  let move_safeness m = match info m with
    { left = l; right = r; } ->
      let ln = info l and rn = info r in
      
      if Args.verbose > 0 then 
	Printf.fprintf stderr "Studying possibility of move between %s and %s\n" 
	  (node_name l) (node_name r);
      if ln.color <> None && rn.color <> None then constrained
      else if l = r then fusionedm
      else if (List.mem r ln.adj) then constrained
      else if (sorted_list_inclusion ln.adj rn.adj) 
	  || (sorted_list_inclusion rn.adj ln.adj) then safe
      else
	let combinedweight = 
	  calc_degree (List.filter (fun n -> n <> l && n <> r) (union_sorted rn.adj ln.adj))
	in
	begin
	  if Args.verbose > 0 then 
	    Printf.fprintf stderr "   -> Combined weight : %i\n" combinedweight;
	  if combinedweight < numfast then safe
	  else if combinedweight < numslow then slowsafe
	  else unsafe
	end
  in

  let add_move m = match info m with
    { left = l; right = r } ->
      (info l).moves <- m::(info l).moves;
      (info r).moves <- m::(info r).moves;
      move m (move_safeness m)
  in

(*  let classify_node no = 
    (*Printf.eprintf "Classifying node %s\n" (Codegen.namer (info no).temp);*)
    let n = info no in (* We do an excessive estimate here because of the precolored nodes *)
    sort_node_adj n;
    n.degree <- List.fold_left 
      (fun i n -> 
	if (belong n lowdensity) || (belong n meddensity) || (belong n highdensity) || 
	  (belong n precolored) || (belong n colored) then begin
	    match (info n).color with
	      None -> i + 1
	    | Some c -> if List.mem c precol then i + 1 else i
	  end else i)
	0 n.adj;
    move no (nodeclass no)
  in*)

(*  let rec find_node n = match info n with
    | { alias = None } -> n
    | { alias = n1 } as n2 -> n2.color = (find_color n1); move n2 colored; n2.color
  in *)

(*  let rec find_node_color n = match info n with
      { color = c } -> c
    | { alias = None } -> (info n).color
    | { alias = n1 } as n2 -> n2.color = (find_color n1); move n2 colored; n2.color
    in*)

  let rec color () =
    if not (empty safe) then
      fusion (pick safe);
    if not (empty lowdensity) then
      simplify (pick lowdensity)
    else if not (empty slowsafe) then
      fusion (pick slowsafe)
    else if not (empty meddensity) then
      simplify (pick meddensity)
    else if not (empty highdensity) then
      raise (Failure "Can't color graph!")

  and setcol n c = 
    n.color <- c;
    match c with
      Some c ->
	if Args.verbose > 0 then
	  Printf.fprintf stderr "Register %s allocated to %s.\n" (temp_name n.temp) (temp_name c)
    | None ->
	if Args.verbose > 0 then
	  Printf.fprintf stderr "Register %s allocated to no color.\n" (temp_name n.temp)
  and simplify n = 
    if Args.verbose > 0 then
	prerr_string ("Simplify "^node_name n^"\n");
    move n stacked;
    List.iter 
      (function no -> 
	let n = info no in 
	n.degree <- n.degree - 1;
	reclassify no) (info n).adj;
    List.iter (function m -> move m frozen) (info n).moves;
    (* print_moves movepartition; *)
    color ();
    move n colored;
    let mycol = 
      try
	List.hd (List.fold_left 
		   (fun l n -> match (info n).color with
		     Some c -> List.filter (function col -> col <> c) l
		   | None -> l) 
		   register_picklist (info n).adj)
      with
	Failure "hd" -> raise (Failure "Can't color node that I expected to be able to color.")
    in
    setcol (info n) (Some mycol)
  
  and fusion m = 
    move m fusionedm;
    let ln = (info m).left and rn = (info m).right in
    if Args.verbose > 0 then
      prerr_string ("Move "^node_name ln^" -> "^node_name rn^"\n");
    move ln fusioned;
    if (info ln).color <> None then begin
      if (info rn).color <> None then raise (Failure "Alloc.alloc tried to coalesce two colored nodes.");
      setcol (info rn) (info ln).color; 
      move rn colored;
    end;
    (info ln).alias <- Some rn;
    let adjust_node_for_alias n = 
      let ni = info n in
      if List.mem rn ni.adj then (ni.degree <- ni.degree - 1; reclassify n)
      else add_interference n rn
    in
    let adjust_move_for_alias m =
      if (belong m safe) || (belong m slowsafe) || (belong m unsafe) then begin
	let mi = info m in
	(* Printf.eprintf "%s %s %b //" (temp_name (info mi.left).temp) (temp_name (info mi.right).temp) (mi.left = mi.right); *)
	if mi.left = ln then begin
	  mi.left <- rn;
	  (info rn).moves <- m::(info rn).moves;
	  (*(info rn).adj <- union_sorted (info ln).adj (info rn).adj*)
	end;
	if mi.right = ln then begin
	  mi.right <- rn;
	  (info rn).moves <- m::(info rn).moves;
	  (* (info rn).adj <- union_sorted (info ln).adj (info rn).adj
	  (info mi.left).adj <- union_sorted (info ln).adj (info rn).adj *)
	end;
	move m (move_safeness m);
      end;
    in
    List.iter adjust_node_for_alias (info ln).adj;
    List.iter adjust_move_for_alias (info ln).moves;
    List.iter 
      (function m -> if (belong m safe) || (belong m slowsafe) || (belong m unsafe) then
	move m (move_safeness m)) 
      (info rn).moves;
    classify_node rn;
    if Args.verbose > 1 then begin
      print_moves movepartition; 
      print_interference nodepartition;
    end;
    color ();
    setcol (info ln) (info rn).color;
    move ln colored
  in

  let getcol = function 
      t -> match (info (findnode t)).color with 
	Some c -> c
      | None -> raise (Failure "Uncolored node!")
  in
  
  flush stdout;
  interference (); (* Prepare interference graph *)
  List.iter add_occurence code; (* Set occurence counts *)
  while not (empty initial) do classify_unsorted_node (pick initial) done; (* Prepares degree info *)
  while not (empty initialm) do add_move (pick initialm) done; (* Add moves to nodes *)
  if Args.verbose > 1 then begin
    print_interference nodepartition;
    print_moves movepartition;
  end;
  color ();
  if Args.verbose > 1 then begin
    print_interference nodepartition;
  end;
  flush stderr;  
  if Args.verbose > 0 then
    List.iter (Codegen.print_finalized_code getcol) code;
  List.flatten (List.map (Codegen.finalize_code getcol) code) 
;;
