(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Pp;;
open Format;;
open Lexer;;

let rec type_expr = function
    Byte -> "byte"
  | SInteger t -> sprintf "sint%d" t
  | UInteger t -> sprintf "uint%d" t
  | Boolean -> "boolean"
  | TConst bigint -> "const" 
  | Void -> "void"
(*  | Array t -> printf "Array (%a)" (out type_expr) t *)
;;
 

let list sep f = function
    [] -> ()
  | h::t -> f h; List.iter (fun x -> printf sep; f x) t
;;

let slist sep f = function
    [] -> ("")
  | h::t -> (f h)^(List.fold_left (fun str x -> str^sep^(f x)) "" t)
;;

let prototype ident arglist = 
  (Lexer.symbol_of_int (snd ident))^"("^(slist ", " (function a -> type_expr (snd a)) arglist)^")";;

let option f = function None -> () | Some x -> f x;;


let var (s, t) = 
  printf "%d : " s;
  printf "%s" (type_expr t);;
    

let var_list l =
  list ", " var l;;

let declarations l = 
  List.iter (fun x -> printf "var "; var x; printf ";@ ") l;;

let sbinop op =
    begin match op with
    | Plus -> "+"
    | Minus -> "-"
    | Times -> "*"
    | Div -> "/"
    | Mod -> "%"
    | Lt -> "<"
    | Gt -> ">"
    | Le -> "<="
    | Ge -> ">="
    | Eq -> "="
    | Ne -> "<>"
    | Xor -> "XOR"
    | Or -> "OR"
    | And -> "AND"
    | L_Or -> "L_OR"
    | L_And -> "L_AND"
    | Concat -> "CONCAT"
    end
;;

let binop op =
  printf "%s" (sbinop op);;

let unop op =
  printf "%s"
    begin match op with
      | Neg -> "Neg"
      | Not -> "Not"
      | Shr -> "Shr"
      | Shl -> "Shl"
      |	Rol -> "Rol"
      | Ror -> "Ror"
      | Ro4 -> "Ro4"
      | Cast t -> sprintf "Cast_%s" (type_expr t)
      | Extract (a,b) -> sprintf "Extract[%d..%d]" a b
      | Extractbit a -> sprintf "Extractbit[%d]" a
      | SAssign a -> sprintf "SAssign %s " (match a with ERA -> "!RA"|ERB -> "!RB"|ERC ->"!RC"|OPTION ->"OPTION")
    end
;;
(*
let priority = function
    Plus | Minus -> 2
  | Times | Div -> 3
  | _ -> 1;;

let lowest = 0;;
let highest = 4;;

let rec expr p = function
    Int n ->
      printf "%s" (string_of_big_int n)
  | Bool b ->
      printf "%s" (if b then "true" else "false")
  | Bin (op, e1, e2) ->
      let p_op = priority op in
      let b = p > p_op in
      let p' = if b then lowest else p_op in
      if b then printf "(";
      expr p' e1;
      binop op;
      expr p' e2;
      if b then printf ")";
  | Get s ->
      printf "%s" s; 
  | Function_call (s, l) ->
      printf "%s(%a)" s (fun out l -> list ", " (expr lowest) l) l; ()
  | Geti (e1, e2) -> 
      expr highest e1;
      printf "[";
      expr lowest e2;
      printf "]";
  | Alloc (e, t) ->
      printf "alloc@;(";
      expr lowest e;
      printf " :@;";
      type_expr t;
      printf ")";
;; 

let expression p e =
  printf "@[";
  expr p e;
  printf "@]";;
  
let rec instruction = function
  | Set (s,e) -> 
      printf "%s := %a" s (fun out -> expression lowest) e;
  | Sequence [ i ] ->
      instruction i;
  | Sequence l ->
      printf "@[<hv>be@[<hv>gin@ ";
      list ";@ " instruction l;
      printf "@]@ end@]";
  | If (e, t, f) -> 
      printf "@[<hv>@[<hv>if@[<hv>@ ";
      expression lowest e;
      printf "@]@ th@[<hv>en@ ";
      instruction t;
      printf "@]@]@ el@[<hv>se@ ";
      instruction f;
      printf "@]@]";
  | While (e, i) -> 
      printf "wh@[ile@;";
      expression lowest e;
      printf " do@ ";
      instruction i;
      printf "@]";
      
  | Procedure_call (s, l) -> 
      printf "%s (" s;
      list ", " (expression lowest) l;
      printf ")";
  | Seti (e1, e2, e3)-> 
      expression highest e1;
      printf "[";
      expression lowest e2;
      printf "] := ";
      expression lowest e3;
  | Write_int n ->
      instruction (Procedure_call ("write", [n]));
  | Writeln_int n ->
      instruction (Procedure_call ("writeln", [n]));
  | Read_int x ->
      instruction (Procedure_call ("read", [Get x]));
;;


let definition (s,f) =
  printf "@[<v>@[<hv 2>%s %s@ "
    (match f.result with None -> "procedure" | _ -> "function")
    s;
  printf "(@[";
  list ",@ " (fun x -> printf "var "; var x) f.arguments;
  printf ")@]";
  option (fun t -> (printf "@ : "; type_expr t)) f.result;
  printf ";@]@ ";
  declarations f.local_vars;
  instruction f.body;
  printf ";@ @ ";;
*)

let rec varexpr ta =
  printf "var(%d:%s " (snd (fst ta)) (symbol_of_int  (snd (fst ta)))  ;
  printf "%s" (type_expr (snd ta));
  printf ")";

and p_expression tp =
  match (snd tp) with
    Int n ->
      printf "%s " (Big_int.string_of_big_int n)
  | Bool b ->
      printf "%s " (if b then "true" else "false")
  | Bin (op, e1, e2) ->
      printf "(";
      binop op;
      printf " ";
      p_expression e1;
      printf " , ";
      p_expression e2;
      printf ") ";
  | Un (op, e1) ->
      printf "(";
      unop op;
      printf " ";
      p_expression e1;
      printf ") "; 
  | Get s ->
      printf "%d:%s " (snd s) (symbol_of_int (snd s)); 
  | Function_call (s,el) ->
      printf "FunCall %d:%s (" (snd s) (symbol_of_int (snd s));
      List.iter p_expression el;
      printf ") ";
  | ESet (s,e1) ->
      printf "ESet(%d:%s," (snd s) (symbol_of_int (snd s));
      p_expression e1;
      printf ") ";
  | Incr (pre,sign,id) ->
      printf "Incr(%s%s,%d:%s)" (if pre==0 then "post" else "pre") (if sign<0 then "decr" else "incr") (snd id) (symbol_of_int (snd id));
  | Ifte (a,b,c) -> iifte (a,b,c);
  

and bloc p = 
  printf "@ @[<v 2>{";
  List.iter p_instruction p;
  printf "@]@ }";

and iif (a,b,c) =
  printf "@[<v>@[<v 4>If @ ";
  p_expression a;
  printf "@]@ @[<v 4>) THEN (";
  p_instruction b;
  printf "@]@ @[<v 4>) ELSE (";
  p_instruction c;
  printf "@]@]@ )";

and iifte (a,b,c) =
  printf "@[<v>@[<v 4>Ifte @ ";
  p_expression a;
  printf "@]@ @[<v 4>) THEN (";
  p_expression b;
  printf "@]@ @[<v 4>) ELSE (";
  p_expression c;
  printf "@]@]@ )";

and fo (a,b,c,d) =
  printf "@[<v>@[<v 4>For INIT (";
  List.iter p_instruction a;
  printf "@]@ @[<v 4>) TEST (@ ";
  p_expression b;
  printf "@]@ @[<v 4>) ITER (";
  List.iter p_instruction c;
  printf "@]@ @[<v 4>) CODE (";
  p_instruction d;
  printf "@]@]@ )";   
and switch (a,b) =
  printf "@[<v>@[<v 4>Switch @ ";
  p_expression a;
  printf "@]@ @[<v 4>) {";
  List.iter switch_statement b;
  printf "@]@]@ }";
    
and switch_statement a =
  printf "@[<v 4> CASE %s : " 
    begin match (snd a.case) with 
     CInt c-> (Big_int.string_of_big_int c)
      | CBool d-> (if d then "true" else "false")
      | Default -> "DEFAULT"
    end;
  p_instruction a.inst;
  
and prcdefinition pa =
  printf "FUN %d:%s ARGS (" (snd pa.name) (symbol_of_int  (snd  pa.name));
  List.iter varexpr pa.arguments;
  printf(")\n");
  p_instruction pa.body;
and p_instruction inst=
  printf "@ l%d " (fst inst).line;
  match (snd inst) with
    Pp.IExpression p -> p_expression p
  | Pp.Bloc p -> bloc p 
  | Pp.Sequence p ->printf "[";List.iter p_instruction p;printf "@ ]";
  | Pp.For (a,b,c,d) -> fo (a,b,c,d)
  | Pp.If (a,b,c) -> iif (a,b,c);
  | Pp.VarExpr (a,b) -> varexpr a;if b then printf "static" ;
  | Pp.ProcDefinition a -> prcdefinition a
  | Pp.Switch (a,b) -> switch (a,b)
  | Pp.Continue -> printf "CONTINUE";
  | Pp.Break -> printf "BREAK";
  | Pp.Return (Expression a) -> printf "RETURN (";p_expression a;printf ")"; 
  | Pp.Return (NoneExpr) -> printf "RETURN";
  | NoneInstruction -> printf "";
;;

let program p =
  printf "@[<v 2>";
  List.iter p_instruction p;
  print_newline();
  printf "@]";
;;
  
