(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Tracetype;;

type banknode = 
    {
    mutable presuccs : (node * int) list;
    mutable succs : banknode list;
    bankinternal : int option;
    mutable bankin : int eoption;
    mutable bankout : int eoption;
    node : node;
    callnum : int;
  } 

let output_bnode chan n =
  let string_of_int_bank_info = function
      None -> "none"
    | Some n -> Printf.sprintf "0x%02X" n
  in
  let string_of_bank_info = function
      ENone -> "none"
    | EMany -> "many"
    | ESome n -> Printf.sprintf "0x%02X" n
  in
  Printf.fprintf chan 
    "%s.%i -> int : %s in : %s out : %s succs : " 
    (Gen.label_string n.node.label) n.callnum
    (string_of_int_bank_info n.bankinternal)
    (string_of_bank_info n.bankin)
    (string_of_bank_info n.bankout);
  List.iter 
    (function n -> Printf.fprintf chan "%s.%i " (Gen.label_string n.node.label) n.callnum) 
    n.succs;
  Printf.fprintf chan "\n"

let output_bnodes chan n = List.iter (output_bnode chan) n

let simplify_banks { nodes = nodes; funcs = funcs } =

  let trychange_bankout dest src =
    if dest.bankinternal <> None || dest.bankout = src then false 
    else (dest.bankout <- src; true)
  in
  let update_successor src dest = 
    (match src.bankout, dest.bankin with
    ENone, _ | _, EMany -> false
  | ESome sb, ENone -> dest.bankin <- ESome sb; true
  | ESome sb, ESome db when sb = db -> false
  | ESome _, ESome _
  | EMany, _ -> dest.bankin <- EMany; true) &&
    (trychange_bankout dest dest.bankin)
  in    
  let update n = (List.filter (update_successor n) n.succs) in

  let function_callers_hash = Hashtbl.create 23 in
  let function_callers n = Hashtbl.find function_callers_hash n.label in
  let add_caller f c = 
    try let old = function_callers f in old := c::!old 
    with Not_found -> Hashtbl.add function_callers_hash f.label (ref [c])
  in

  let bnode_of_node_hash = Hashtbl.create 23 in
  let bnode_of_node = Hashtbl.find bnode_of_node_hash in
  let banknodes = ref [] in
  let make_bnodes n = 
    let new_banknode n c b s =
      let newbnode = { node = n; callnum = c; bankinternal = b; 
		       presuccs = s; succs = []; bankin = ENone; 
		       bankout = match b with Some b -> ESome b | None -> ENone; }
      in
      Hashtbl.add bnode_of_node_hash (n, c) newbnode;
      banknodes := newbnode::!banknodes
    in 
    let make_bnode ((cbank, i) as param) = function 
	Bank b -> Some b, i
      | Call dest -> new_banknode n i cbank [dest.dest, 0]; None, i + 1
      | _ -> param
    in
    let b, i = List.fold_left make_bnode (None, 0) n.code in
    if n.lastinstr = Ret then 
      new_banknode n i b 
	(match n.belongfunc with
	  None -> failwith "Ret from a non-function detected in bank.ml"
	| Some f -> List.map (function n, i -> n, i + 1) f.callers)
    else new_banknode n i b (List.map (fun n -> n, 0) (successors n));
  in
  let fill_succs bn = bn.succs <- List.map bnode_of_node bn.presuccs in
      
  List.iter make_bnodes nodes;
  let bnodes = List.rev !banknodes in
  List.iter fill_succs bnodes;
  if Args.verbose > 0 then begin
    print_string "\nPrinting bank node graph.*************************************************\n";
    output_bnodes stdout bnodes;
    print_string "\nEnd of bank node graph.*************************************************\n";
    fixpoint update bnodes;
    print_string "\nPrinting processed bank node graph.*************************************************\n";
    output_bnodes stdout bnodes;
    print_string "\nEnd of processed bank node graph.*************************************************\n";
  end
;;

