(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Code;;

type bitreg = Bit of Gen.temp * int;;

let code = ref [];;
let emit c = code := c::!code;;

let r = Gen.registers;;
let regINDF = r.(0);;
let regW = r.(1);;
let regPC = r.(2);;
let regSTATUS = r.(3);;
let regFSR = r.(4);;
let regRA = r.(5);;
let regRB = r.(6);;
let regRC = r.(7);; 

let regM = r.(256);;
let regOPTION = r.(257);;
let regERA = r.(258);;
let regERB = r.(259);;
let regERC = r.(260);;
let regC = r.(261);;
let regDC = r.(262);;
let regZ = r.(263);;
let regPD = r.(264);;
let regTO = r.(265);;
let regPA0 = r.(266);;
let regPA1 = r.(267);;
let regPA2 = r.(268);;
let regC = r.(261);;
let regDC = r.(262);;
let regZ = r.(263);;
let regPD = r.(264);;
let regTO = r.(265);;
let regPA0 = r.(266);;
let regPA1 = r.(267);;
let regPA2 = r.(268);;

let regCbit = Bit (regC, 0);;
let regDCbit = Bit (regDC, 1);;
let regZbit = Bit (regZ, 2);;

let always_used_registers = 
  [regERA; regERB; regERC; regOPTION; regM; regFSR; regRA; regRB; 
   regRC; regPC];;

let name_of_register = [|
  "INDF"; "WW"; "PC"; "STATUS"; "FSR"; "RA"; "RB"; "RC"
|];;

let name_of_ext = [|
  "M"; "!OPTION"; "!RA"; "!RB"; "!RC"; "C"; "DC"; "Z"; "PD"; "TO"; "PA0"; "PA1"; "PA2" 
|];;

let defby = function
    Ass.Oper (_, _, d, _, _) -> d
  | Ass.Move (_, d) -> [d; regW; regZ]
  | Ass.Label _ -> [];;

let usedby = function
    Ass.Oper (_, s, _, _, _) -> s
  | Ass.Move (s, _) -> [s]
  | Ass.Label _ -> [];;

let move_parasite_dests = [regW];;

let namer r =
  let rec tryarrays i = function
      [] -> if i < 255 then Printf.sprintf "r0x%02X" i else Gen.namer r
    | (a, ofs)::t -> try a.(i - ofs) with Invalid_argument _ -> tryarrays i t
  in
  tryarrays (Gen.temp_int r) [name_of_register, 0; name_of_ext, 256];;

let reg_bank r = 
  if (r >= 256) or (r land 16 = 0) then 
    raise (Failure ("Shouldn't generate bank for register"^(namer r)));
  r land 0xE0

let registers = 
  let rec gen = function 269 -> [] | n -> n::(gen (n + 1)) in 
  List.filter (function n -> n < 32 or n >= 256 or n mod 32 >= 16) (gen 0);;

let fast_registers = regW::(List.filter (function i -> i >= 8 && i < 16) registers);;
let slow_registers = List.filter (function i -> i >= 16 && i < 256) registers;;

exception CodeGenError of string;;

let listextract i j l = 
  let rec extract n = function
      l when n = j + 1 -> []
    | [] -> invalid_arg "listextract"
    | h::t -> let r = extract (n + 1) t in 
      if n >= i then h::r else r
  in
  extract 0 l;;

let rec list_last = 
  function [] -> raise (CodeGenError "Empty integer type.") | [a] -> a | h::t -> list_last t;;

let gen_bit t i = 
  let b = try List.nth t (i / 8) with 
    _ -> (raise (CodeGenError "Codegen.emit_unop Extractbit on invalid bit.")) 
  in
  Bit (b, (i mod 8));;

let emit_oper a s d f l = emit (Ass.Oper (a, s, d, f, l));;
let emit_mov d s = if s <> d then emit (Ass.Move (s, d));;

(*let rec emit_mov (d : Gen.temp) (s : Gen.temp) = match s, d with
  s, d when s = d -> ()
| s, d when s = regW -> emit_move "mov ^d0, W ; ^s0" regW d (* #### modifies Z *)
| s, d when d = regW -> emit_move "mov W, ^s0 ; ^d0" s regW
| s, d -> emit_move "mov W, ^s0\n"^ident^"mov ^d0, W" s regW*)

let emit_label l =
  emit (Ass.Label ((Gen.label_string l), l));;

let emit_clr d =
  emit_oper "clr ^d0" [] [d; regZ] true [];;

(*let emit_clrifnot d s =
  let l = Gen.new_label () in
  emit_oper "sb ^s0" [s] [] true [l];
  emit_oper "clrb ^d0" [d] [d] true [];
  emit_label l

let emit_clrif d s =
  let l = Gen.new_label () in
  emit_oper "snb ^s0" [s] [] true [l];
  emit_oper "clrb ^d0" [d] [d] true [];
  emit_label l

let emit_setifnot d s =
  let l = Gen.new_label () in
  emit_oper "sb ^s0" [s] [] true [l];
  emit_oper "setb ^d0" [d] [d] true [];
  emit_label l

let emit_setif d s =
  let l = Gen.new_label () in
  emit_oper "snb ^s0" [s] [] true [l];
  emit_oper "setb ^d0" [d] [d] true [];
  emit_label l *)

let emit_setb d = 
  emit_oper "setb ^d0" [] [d] true [];;

let emit_clrb d =
  emit_oper "clrb ^d0" [] [d] true [];;

let emit_stc () = 
  emit_oper "setb C" [] [regC] true [];;

let emit_clc () =
  emit_oper "clrb C" [] [regC] true [];; 

let emit_movc d i =
(*  if i != 0 then begin*)
    emit_oper ("mov W, #"^(string_of_int i)) [] [regW] true [];
    emit_mov d regW 
(*  end else
    emit_clr d;;*)

let emit_addc d s1 s2 =
  if (s2 = regW && s1 <> regW) then
    raise (CodeGenError "emit_addc called with W-conflicting parameters.");
  emit_mov regW s1;
  emit_oper "add W, ^s0" [s2; regW; regC] [regW; regC; regDC; regZ] true [];
  emit_mov d regW;;

let emit_add d s1 s2 =
  emit_clc ();
  emit_addc d s1 s2;;

let emit_subc d s1 s2 =
  if (s1 = regW && s2 <> regW) then
    raise (CodeGenError "emit_addc called with W-conflicting parameters.");
  emit_mov regW s2;
  emit_oper "mov W,^s0-W" [s1; regW; regC] [regW; regC; regDC; regZ] true [];
  emit_mov d regW;;

let emit_subcfrom d s =
  emit_mov regW s;
  emit_oper "sub ^s0, W" [d; regW; regC] [d; regC; regDC; regZ] true [];;

let emit_sub d s1 s2 =
  emit_stc ();
  emit_subc d s1 s2;;

let emit_and d s1 s2 =
  if (s2 = regW && s1 <> regW) then
    raise (CodeGenError "emit_addc called with W-conflicting parameters.");
  emit_mov regW s1;
  emit_oper "and W, ^s0" [s2; regW] [regW; regZ] true [];
  emit_mov d regW;;

let emit_or d s1 s2 =
  if (s2 = regW && s1 <> regW) then
    raise (CodeGenError "emit_addc called with W-conflicting parameters.");
  emit_mov regW s1;
  emit_oper "or W, ^s0" [s2; regW] [regW; regZ] true [];
  emit_mov d regW;;

let emit_xor d s1 s2 =
  if (s2 = regW && s1 <> regW) then
    raise (CodeGenError "emit_addc called with W-conflicting parameters.");
  emit_mov regW s1;
  emit_oper "xor W, ^s0" [s2; regW] [regW; regZ] true [];
  emit_mov d regW;;

(*let emit_movbc d c =
  (if c then emit_setb else emit_clrb) d;; *)

let emit_not d s =
  emit_mov d s;
  emit_oper "not ^s0" [d] [d; regZ] true [];;

let emit_shr d s =
  emit_oper "mov W,>> ^s0" [s; regC] [regW; regC] true [];
  emit_mov d regW;;

let emit_shl d s =
  emit_oper "mov W,<< ^s0" [s; regC] [regW; regC] true [];
  emit_mov d regW;;

let emit_getlsb d s =
  emit_oper "mov W,>> ^s0" [s] [regW; regC] true [];
;;

let emit_getmsb d s =
  emit_oper "mov W,<< ^s0" [s] [regW; regC] true [];
;;

let emit_notzero s =
  emit_oper "test ^s0" [s] [regZ] true [];;

let emit_stillnotzero s =
  emit_oper ("snb Z\n"^Ass.indent^"test ^s0") [s; regZ] [regZ] true [];;

let emit_swapnibble d s =
  emit_mov d s;
  emit_oper "swap ^s0" [d] [d] true [];;

let emit_setbit (Bit (t, i)) =
  emit_oper ("sb Z\n"^Ass.indent^"clrb ^d0."^(string_of_int i)^
	     "\n"^Ass.indent^"snb Z\n"^Ass.indent^"setb ^d0."^(string_of_int i)) [t; regZ] [t] true [];;

(*let emit_notb d s =
  let l = Gen.new_label () in
  emit_oper "snb ^s0" [s] [] true [];
  emit_oper "clrb ^d0" [] [d] true [];
  emit_oper "sb ^s0" [s] [] true [];
  emit_oper "setb ^d0" [] [d] true [];
  emit_label l;;*)

let emit_decWif (Bit (r, b)) =
  emit_oper ("snb ^s0."^(string_of_int b)^"\n"^Ass.indent^"dec WW") [r; regW] [regW] true [];;

let emit_incWif (Bit (r, b)) =
  emit_oper ("snb ^s0."^(string_of_int b)^"\n"^Ass.indent^"inc WW") [r; regW] [regW] true [];;

let emit_incWifnot (Bit (r, b)) =
  emit_oper ("sb ^s0."^(string_of_int b)^"\n"^Ass.indent^"inc WW") [r; regW] [regW] true [];;

let emit_extractbittoW t =
  emit_movc regW 0;
  emit_incWif t;;

let emit_extractnotbittoW t =
  emit_movc regW 1;
  emit_decWif t;;

let extractbit s =
  let d = Gen.new_temp () in
  emit_extractbittoW s;
  emit_mov d regW;
  [d];;

let extractnotbit s =
  let d = Gen.new_temp () in
  emit_extractnotbittoW s;
  emit_mov d regW;
  [d];;

let emit_jump l = 
  emit_oper ("jmp "^(Gen.label_string l)) [] [] false [l];;
    (* Big problem if not in same page *)

let emit_test t =
  emit_oper "test ^s0" [t] [regZ] true []
    
let emit_jcond t b l1 l2 = 
  let l1' = if b then l1 else l2 in
  let l2' = if b then l2 else l1 in
  emit_test t;
  emit_oper ("jcond Z, "^
	     (Gen.label_string l2')^", "^
	     (Gen.label_string l1')) 
    [regZ] [] false [l2; l1];;

let emit_jcondb (Bit (t, i)) b l1 l2 = 
  let l1' = if b then l1 else l2 in
  let l2' = if b then l2 else l1 in
  emit_oper ("jcond ^s0."^(string_of_int i)^", "^
	     (Gen.label_string l1')^", "^
	     (Gen.label_string l2')) 
    [t] [] false [l1; l2];;

let emit_signedcomparetoW t1 t2 =
  let msb1 = list_last t1 in
  let msb2 = list_last t2 in
  let highbyte = Gen.new_temp () in
  emit_clr regW;
  emit_incWif (Bit (msb1, 7));
  emit_incWif (Bit (msb2, 7));
  emit_mov highbyte regW;
  List.iter2 (emit_subc regW) t1 t2;
  emit_clr regW;
  emit_addc regW regW highbyte;
  let out = Gen.new_temp () in
  emit_mov out regW;
  out
;;

let emit_movspecialfromW sreg =
  let sreg_name = 
    if sreg = regOPTION then "!OPTION"
    else if sreg = regERA then "!RA"
    else if sreg = regERB then "!RB"
    else if sreg = regERC then "!RC"
    else failwith "Invalid special register in Codegen.emit_movespecialfromW"
  in
  emit_oper ("mov "^sreg_name^", W") [regW] [sreg; regW] true []

(*let emit_un_lchain initfn seqfn = function
    [] -> ()
  | h::t -> IntTemp ((initfn h)::(List.map seqfn t));;*)

let rec emit_exp = function
    Const c ->
      List.map (function i -> let t = Gen.new_temp () in emit_movc t i; t) c
  | Un (op, e) -> emit_unop (emit_exp e) op
  | Bin (op, e1, e2) -> emit_binop (emit_exp e1) (emit_exp e2) op
  | Temp t -> t
  | Call (f, pl) ->
      (try
	List.iter2 
	  (fun s d -> 
	    ignore (emit_exp (Bin (Assign, Temp d, s))))
            pl f.args
      with
	Invalid_argument _ -> raise (CodeGenError "Wrong number of arguments in Codegen.emit_exp"));
      emit_oper 
	("call "^(Gen.label_string f.name)) f.uses f.defs true [];
      f.result	
(*  | Ifte (Un (Extractbit i, e), t, f) ->
      let c = emit_exp e in
      let ltrue = Gen.new_label () 
      and lfalse = Gen.new_label () 
      and ldone = Gen.new_label () in
      emit_jcondb (gen_bit c i) true ltrue lfalse;
      emit_label ltrue;
      let ttrue = emit_exp t in
      emit_jump ldone;
      emit_label lfalse;
      let tfalse = emit_exp f in
      List.iter2 emit_mov ttrue tfalse;
      emit_jump ldone;
      emit_label ldone;
      ttrue
  | Ifte (c, t, f) ->
      let ct = emit_exp c in
      let ltrue = Gen.new_label () 
      and lfalse = Gen.new_label () 
      and ldone = Gen.new_label () in
      (match ct with
	[t] -> emit_jcond t true ltrue lfalse;
      |	_ -> raise (CodeGenError "?: with non bool expression in Codegen.emit_stm."));
      emit_label ltrue;
      let ttrue = emit_exp t in
      emit_jump ldone;
      emit_label lfalse;
      let tfalse = emit_exp f in
      List.iter2 emit_mov ttrue tfalse;
      emit_jump ldone;
      emit_label ldone;
      ttrue*)
  | Ifte _ -> raise (CodeGenError "Found an Ifte remaining in intermediate code in Codegen.emit_exp.")
  | Name _ -> raise (CodeGenError "Found a Name remaining in intermediate code in Codegen.emit_exp.")
(*  | _ -> raise (CodeGenError "Unmatched expression type in Codegen.emit_exp")*)

and emit_unop t = 
  let rec rmap f = function [] -> [] | h::t -> (f h)::(rmap f t) in
  let imap f s =
    (List.map (function s -> let d = Gen.new_temp () in f d s; d) s)
  in
  let revimap f s =
    (rmap (function s -> let d = Gen.new_temp () in f d s; d) s)
  in function
      Extract (i, j) -> listextract i j t
    | Extractbit i ->
	extractbit (gen_bit t i);
    | Neg ->
	emit_stc ();
	imap (fun d s -> emit_clr d; emit_subcfrom d s) t
    | Not ->
	imap emit_not t
    | Shr ->
	emit_clc ();
	revimap emit_shr t
    | Shl ->
	emit_clc ();
	imap emit_shl t
    | Asr ->
	let msb = list_last t in
	emit_getmsb (Gen.new_temp ()) msb; (* To get msbit *)
	revimap emit_shr t
    | Ro4 ->
	(match t with
	  [t] -> 
	    let t2 = Gen.new_temp () in
	    emit_swapnibble t2 t; [t2]
	| _ -> raise (CodeGenError "Rot4 on multi-byte integer"))
    | Ror ->
	(try
	  let lsb = List.hd t in
	  emit_getlsb (Gen.new_temp ()) lsb; (* to get lsbit *)
	  imap emit_shr t
	with _ -> raise (CodeGenError "Empty integer temporary while generating Ror."))
    | Rol ->
	let msb = list_last t in
	emit_getmsb (Gen.new_temp ()) msb; (* To get msbit *)
	revimap emit_shl t
    | UExtend i ->
	emit_clr regW;
	let tl = ref [] in
	for j = 1 to i do
	  let t = Gen.new_temp () in
	  emit_mov t regW;
	  tl := t::!tl
	done;
	t@(!tl)
    | SExtend i ->
	let msb = list_last t in
	emit_clr regW;
	emit_decWif (Bit (msb, 7));
	let tl = ref [] in
	for j = 1 to i do
	  let t = Gen.new_temp () in
	  emit_mov t regW;
	  tl := t::!tl
	done;
	t@(!tl)
    | NotZero ->
	(try
	  emit_notzero (List.hd t);
	  List.iter emit_stillnotzero (List.tl t);
	with _ -> ());
	extractnotbit regZbit;
    | SAssign sreg -> 
	let d = Gen.new_temp () in
	(match t with
	  [t] -> emit_mov regW t;
	| _ -> failwith "Codegen.emit_unop : SAssign on non 8 bit temporary!");
	emit_movspecialfromW sreg;
	emit_mov d regW;
	[d]

and emit_binop t1 t2 = (* Must add mult, div, comp #### *)
  let imap f s1 s2 =
    try
      (List.map2 (fun s1 s2 -> let d = Gen.new_temp () in f d s1 s2; d) s1 s2)
    with 
      Invalid_argument _ -> raise (CodeGenError "Mismatched integer widths for binary operator.")
  in function
      Plus ->
	emit_clc ();
	imap emit_addc t1 t2
    | Minus ->
	emit_stc ();
	imap emit_subc t1 t2
    | And ->
	imap emit_and t1 t2
    | Or ->
	imap emit_or t1 t2
    | Xor ->
	imap emit_xor t1 t2
    | Eq ->
	emit_stc ();
	List.iter2 (emit_subc regW) t1 t2;
	extractbit regZbit;
    | Ne ->
	emit_stc ();
	List.iter2 (emit_subc regW) t1 t2;
	extractnotbit regZbit;
    | ULt ->
	emit_stc ();
	List.iter2 (emit_subc regW)  t1 t2;
	extractnotbit regCbit;
    | UGe ->
	emit_stc ();
	List.iter2 (emit_subc regW)  t1 t2;
	extractbit regCbit;
    | UGt ->
	emit_stc ();
	List.iter2 (emit_subc regW)  t2 t1;
	extractnotbit regCbit;
    | ULe ->
	emit_stc ();
	List.iter2 (emit_subc regW)  t2 t1;
	let t = Gen.new_temp () in
	extractbit regCbit;
    | SLe ->
	let r = emit_signedcomparetoW t1 t2 in
	extractnotbit (Bit (r, 0));
    | SGt ->
	let r = emit_signedcomparetoW t1 t2 in
	extractbit (Bit (r, 0));
    | SGe ->
	let r = emit_signedcomparetoW t2 t1 in
	extractnotbit (Bit (r, 0));
    | SLt ->
	let r = emit_signedcomparetoW t2 t1 in
	extractbit (Bit (r, 0));
    | Assignbit i ->
	(match t2 with
	  [t] as rslt ->
	    emit_test t;
	    emit_setbit (gen_bit t1 i);
	    rslt
	| _ -> raise (CodeGenError "Invalid bool temporary with Assignbit in Codegen.emit_bin."))
    | Assign ->
	(try
	  List.map2 (fun d s -> emit_mov d s; d) t1 t2;
	with Invalid_argument _ -> 
	  failwith ("Operand size does not match in Codegen.emit_binop - Assign "^
		    (string_of_int (List.length t1))^" "^(string_of_int (List.length t2))))
    | Concat ->
	t2@t1
    | Comma -> raise (CodeGenError "Comma operator in Codegen") 
(*    | _ -> raise (CodeGenError "Bad binary operator in Codegen") *)

let rec emit_stm = function
    Label l -> emit_label l
  | Seq l -> List.iter emit_stm l
  | Exp e -> ignore (emit_exp e)
  | Jump l -> emit_jump l
  | Cjump (Const c, b, l1, l2) -> 
      emit_jump (if (List.exists (function 0 -> false | _ -> true) c) = b then l1 else l2)
  | Cjump (Un (Extractbit i, e), b, l1, l2) -> 
      let c = emit_exp e in
      emit_jcondb (gen_bit c i) b l1 l2
  | Cjump (e, b, l1, l2) -> 
      (match emit_exp e with 
	[t] -> 
	  emit_jcond t b l1 l2
      |	_ -> raise (CodeGenError "Cjump with non bool expression in Codegen.emit_stm."))
  | Return (f, t) -> 
      ignore (emit_exp (Bin (Assign, Temp f.result, t)));
      emit_oper "ret" 
	(f.result @ f.statics) [] false [];
;;

(*let use_temps = function
    Ass.Oper (_, s, _, _) -> s
  | Ass.Label _ -> []
  | Ass.Move (_, s, _) -> [s];;

let gen_temps = function
    Ass.Oper (_, _, d, _) -> d
  | Ass.Label _ -> []
  | Ass.Move (_, _, d) -> [d];;*)

let emit_code stmlist =
  code := [];
  List.iter emit_stm stmlist;
  let rslt = List.rev !code in
  code := [];
  rslt;;

let print_bank r =
  if List.mem r slow_registers then
    print_string ("bank #"^(string_of_int (reg_bank r))^"\n")
;;
    
let print_instr str s d =
  let i = ref 0 in
  print_string Ass.indent;
  while !i < String.length str do
    if str.[!i] = Ass.escape then begin
      print_string (namer (try
	List.nth 
	  (if str.[!i + 1] = 's' then s
	  else if str.[!i + 1] = 'd' then d
	  else raise (Failure "Internal error : Invalid escape sequence in Codegen.print_instr"))
	  ((int_of_char str.[!i + 2]) - (int_of_char '0'))
      with
	Failure s -> raise (Failure s)
      |	_ -> raise (Failure "Internal error : Problem converting escape sequence in Codegen.print_instr")));
      i := !i + 2;	  
    end else
      print_char str.[!i];
    incr i
  done

let print_bank2 l1 l2 =
  match
    (List.filter (fun r -> List.mem r slow_registers) l1) @
    (List.filter (fun r -> List.mem r slow_registers) l2)
  with
  | [r; s] when r = s -> print_bank r
  | [r] -> print_bank r
  | [] -> ()
  | l -> raise (Failure ("Multiple bank switches for instruction : "^(List.fold_left (fun s r -> s^" "^(namer r)) "" l)))

let print_finalized_code col = function
    Ass.Label (s, _) -> print_string s; print_newline ();
  | Ass.Move (s, d) ->
      let sr = col s and dr = col d in
      if sr = dr then (print_string (Ass.indent^"; dropped mov "^(namer dr)^", "^(namer sr)^" ; "^(namer d)^" <- "^(namer s)^"\n"))
      else if sr = regW then begin
	print_bank dr;
	print_string (Ass.indent^"mov "^(namer dr)^", W ; "^(namer d)^" <- W\n")
      end else if dr = regW then begin
	print_bank sr;
	print_string (Ass.indent^"mov W, "^(namer sr)^" ; W <- "^(namer s)^"\n")
      end else begin
	print_bank sr;
	print_string (Ass.indent^"mov W, "^(namer sr)^" ; W <- "^(namer s)^"\n");
	print_bank dr;
	print_string (Ass.indent^"mov "^(namer dr)^", W ; "^(namer d)^" <- W\n")
      end
  | Ass.Oper (str, s, d, _, _) ->
      let sr = List.map col s in
      let dr = List.map col d in
      if try (String.sub str 0 4) <> "call" with _ -> true then
	print_bank2 sr dr;
      print_instr str sr dr;
      print_newline ()
;;

let print_program_ass p = 
    List.iter (function { Ass.code = c } -> 
      List.iter (function c -> Ass.print stdout namer c; print_newline ()) c) p;;

let gen_bank r =
  if List.mem r slow_registers then
    [Fasm.Bank (reg_bank r)]
  else
    []
;;
    
let gen_bank2 l1 l2 =
  match
    (List.filter (fun r -> List.mem r slow_registers) l1) @
    (List.filter (fun r -> List.mem r slow_registers) l2)
  with
  | [r; s] when r = s -> gen_bank r
  | [r] -> gen_bank r
  | [] -> []
  | l -> raise (Failure ("Multiple bank switches for instruction : "^(List.fold_left (fun s r -> s^" "^(namer r)) "" l)))

let subst_regs str s d =
  let i = ref 0 in
  let outs = ref "" in
  while !i < String.length str do
    outs := !outs^
      (if str.[!i] = Ass.escape then begin
	i := !i + 2;	  
	namer (try
	  List.nth 
	    (if str.[!i - 1] = 's' then s
	    else if str.[!i - 1] = 'd' then d
	    else raise (Failure "Internal error : Invalid escape sequence in Codegen.print_instr"))
	    ((int_of_char str.[!i]) - (int_of_char '0'))
	with
	  Failure s -> raise (Failure s)
	|	_ -> raise (Failure "Internal error : Problem converting escape sequence in Codegen.print_instr"));
      end else
	 String.sub str !i 1);
    incr i
  done;
  !outs;;
    
let finalize_code col = 
  let string_starts_with s1 s2 = 
    try (String.sub s1 0 (String.length s2)) = s2 with _ -> false
  in
  function
      Ass.Label (_, l) -> [Fasm.Label l]
    | Ass.Move (s, d) ->
	let sr = col s and dr = col d in
	if sr = dr then [Fasm.Oper (Ass.indent^"; dropped mov "^(namer dr)^", "^(namer sr)^" ; "^(namer d)^" <- "^(namer s)^"\n", 0)]
	else if sr = regW then begin
	  (gen_bank dr)@
	  [Fasm.Oper (Ass.indent^"mov "^(namer dr)^", W ; "^(namer d)^" <- W", 1)]
	end else if dr = regW then begin
	  (gen_bank sr)@
	  [Fasm.Oper (Ass.indent^"mov W, "^(namer sr)^" ; W <- "^(namer s), 1)]
	end else begin
	  (gen_bank sr)@
	  [Fasm.Oper (Ass.indent^"mov W, "^(namer sr)^" ; W <- "^(namer s), 1)]@
	  (gen_bank dr)@
	  [Fasm.Oper (Ass.indent^"mov "^(namer dr)^", W ; "^(namer d)^" <- W", 1)]
	end
    | Ass.Oper (str, s, d, n, l) ->
	let sr = List.map col s in
	let dr = List.map col d in
	(if try (String.sub str 0 4) <> "call" with _ -> true then
	  gen_bank2 sr dr
	else [])@
	let str = subst_regs str sr dr in
	(match n with
	  false -> 
	    (match l with
	      [l1; l2] ->
		let ex = Failure "Badly formatted jcond in Codegen.finalize_code." in
		if not (string_starts_with str "jcond ") || n then
		  raise ex;
		let i = try String.index str ',' with
		  _ -> raise ex
		in
		[Fasm.Jcond ((String.sub str 6 (i - 6)), l1, l2)]
	    | [l] ->
		let ex = Failure "Badly formatted jmp in Codegen.finalize_code." in
		if not (string_starts_with str "jmp ") || n then
		  raise ex;
		[Fasm.Jmp l]
	    | [] -> 
		let ex = Failure ("Badly formatted ret in Codegen.finalize_code : "^str) in
		if not (string_starts_with str "ret") || n then
		  raise ex;
		[Fasm.Ret]
	    | _ -> raise (Failure "Badly formatted instruction in Codegen.finalize_code."))
	| true ->
	    if l <> [] then
	      raise (Failure "Badly formatted instruction in Codegen.finalize_code.");
	    if string_starts_with str "call " then
	      [Fasm.Call (Gen.label_of_name (String.sub str 5 ((String.length str) - 5)))]
	    else
	      let valid_lines s =
		let validline = ref true in
		let lines = ref 0 in
		for i = 0 to (String.length str) - 1 do
		  match !validline, s.[i] with
		    _, '\n' -> validline := true
		  | true, ' ' -> ()
		  | true, ';' -> validline := false
		  | true, _ -> validline := false; incr lines
		  | false, _ -> ()
		done;
		!lines
	      in
	      [Fasm.Oper (Ass.indent^str, valid_lines str)])
;;
