(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Code

(* Extracts calls from expressions in a statement. *)

let rec (extractcalls_stm : stm -> stm) = function
    Seq sl -> Seq (List.map extractcalls_stm sl)
  | Exp e -> (extractcalls_rootexp e)
  | Cjump (e, b, l1, l2) -> 
      let sl, ex = extractcalls_exp e in 
      Seq [sl; Cjump (ex, b, l1, l2)]
  | Label _ as l -> l
  | Jump _ as j -> j
  | Return (f, e) -> 
      let sl, ex = extractcalls_exp e in 
      Seq [sl; Return (f, ex)]

(*  | s -> s*)

and extractcalls_rootexp ex = match ex with 
  | Call (fr, exl) -> 
      let sl, exl' = List.split (List.map extractcalls_exp exl) in
      Seq [Seq sl; Exp (Call (fr, exl'))]
  | Bin (Assign, exlvl, Call (fr, exl)) -> 
      let sl, exl' = List.split (List.map extractcalls_exp exl) in
      let sllvl, exlvl' = extractcalls_exp exlvl in
      Seq [sllvl; Seq sl; Exp (Bin (Assign, exlvl',Call (fr, exl')))]
  | e -> 
      let sl, ex = extractcalls_exp e in 
      Seq [sl; Exp ex]

and (extractcalls_exp : exp -> stm * exp) = function
    Un (op, ex) -> 
      let sl, ex' = extractcalls_exp ex in 
      sl, Un (op, ex')
  | Bin (op, ex1, ex2) ->
      let sl1, ex1' = extractcalls_exp ex1
      and sl2, ex2' = extractcalls_exp ex2 in
      (match op with
	Comma -> Seq [sl1; Exp ex1'; sl2], ex2'
      |	_ -> Seq [sl1; sl2], Bin (op, ex1', ex2')
	)
  | Call (f, expl) as callexp ->
      let paramassigns = 
	Seq (List.map2
	       (fun exp arg -> 
		 let tmp = List.map (fun _ -> Gen.new_temp ()) arg in
		 let seq, exp = extractcalls_exp exp in
	    	 Seq [seq; Exp (Bin (Assign, Temp tmp, exp))]) 
	       expl f.args)
      in
      let tmp = List.map (fun _ -> Gen.new_temp ()) f.result in
      Seq [paramassigns; Exp (Bin (Assign, Temp tmp, callexp))], Temp tmp
  | Temp _ as t -> Seq [], t
  | Name _ as n -> Seq [], n (* Like this if we add instructions we will get a warning *)
  | Const _ as c -> Seq [], c
  | Ifte (c, e1, e2, i) ->
      let t = Array.to_list (Array.init i (fun _ -> Gen.new_temp ())) in
      let cstm, cexp = extractcalls_exp c in
      let e1stm, e1exp = extractcalls_exp e1 in
      let e2stm, e2exp = extractcalls_exp e2 in
      let ltrue = Gen.new_label () in
      let lfalse = Gen.new_label () in
      let ldone = Gen.new_label () in
      Seq [
      cstm;
      Cjump (cexp, true, ltrue, lfalse);
      Label lfalse;
      e2stm;
      Exp (Bin (Assign, Temp t, e2exp));
      Label ltrue;
      e1stm;
      Exp (Bin (Assign, Temp t, e1exp));
      Label ldone;
    ], Temp t
;;

let flatten_stm s = 
  let rec helper s t = match s with
      Seq s -> List.fold_right helper s t
    | s -> s::t
  in
  List.fold_right helper s [];;

let extractcalls_func = function
    { code = c } as f -> 
      f.code <- flatten_stm  (List.map extractcalls_stm c);;
      

let extractcalls_program = function 
    { main = f; funcs = fl } -> List.iter extractcalls_func (f::fl);;

