(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

type ('o, 'i) t = { 
    move_to_top : 'o -> 'o -> bool; (* says if first element should move nearer to top *)
      index : 'o-> 'i; (* gives an index for an element *)
	hash : ('i, int) Hashtbl.t;
	mutable numelems : int;
	mutable heap : 'o option array 
  }

let l_son i = 2 * i + 1
let r_son i = 2 * i + 2
let father i = (i - 1) / 2

exception No_change
exception Collision
exception Empty
exception Incoherent_heap

let hash_add h o i = Hashtbl.add h.hash (h.index o) i
let hash_remove h o = 
  Hashtbl.remove h.hash (h.index o)

let at_index h i = match h.heap.(i) with
  None -> raise Incoherent_heap
| Some o -> o

let move h i j = (* i is not in hash table *)
  let oj = at_index h j in
  hash_remove h oj;
  hash_add h oj i;
  h.heap.(i) <- Some oj
;;
    
let finalize_percolation h o i =
  hash_add h o i;
  h.heap.(i) <- Some o;
  raise No_change

let should_move h i j =
  h.move_to_top (at_index h i) (at_index h j)

let rec percolate_down h o i =
  let try_swap_down i =
    let l = l_son i and r = r_son i in
    let should_bump_up i =
      h.move_to_top o (at_index h i)
    in
    if r < h.numelems && should_move h l r then 
      if should_bump_up r then
	(move h i r;r)
      else
	finalize_percolation h o i
    else if l < h.numelems && should_bump_up l then
      (move h i l; l)
    else
      finalize_percolation h o i
  in
  try
    percolate_down h o (try_swap_down i)
  with
    No_change -> ()
  
let rec percolate_up h o i =
  let try_swap_up i =
    if i = 0 then finalize_percolation h o i;
    let f = father i in
    let should_bump_down i =
      h.move_to_top (at_index h i) o
    in
    if should_bump_down f then
      (move h i f; f)
    else
      finalize_percolation h o i
  in
  try
    percolate_up h o (try_swap_up i)
  with
    No_change -> ()

let empty h = h.numelems = 0

let create mtt ind i =
  if i > 0 then
    { move_to_top = mtt; index = ind; hash = Hashtbl.create i; 
      heap = Array.create i None; numelems = 0 }
  else
    raise (Invalid_argument "Heap.create")

let remove_i h i =
  h.numelems <- h.numelems - 1;
  let j = h.numelems in
  let oi = at_index h i in
  hash_remove h oi;
  if j <> i then begin
    let oj = at_index h j in
    hash_remove h oj;
    percolate_down h oj i
  end;
  h.heap.(j) <- None

let remove h e =
  remove_i h (Hashtbl.find h.hash (h.index e)) (* Raises Not_found if not found *)

let head h =
  match h.heap.(0) with
    None -> raise Empty
  | Some o -> remove_i h 0; o
  
let find h ind = at_index h (Hashtbl.find h.hash ind)

let collision h o =
  try Some (find h (h.index o)) with Not_found -> None

let increase_heap h =
  let n = Array.length h.heap in
  let nh = Array.create (1 + 3 * n / 2) None in
  Array.blit h.heap 0 nh 0 n;
  h.heap <- nh

let add h o =
  if collision h o <> None then raise Collision;
  let i = h.numelems in 
  h.numelems <- h.numelems + 1;
  if i >= Array.length h.heap then 
    increase_heap h;
  percolate_up h o i

(*let testheap h =
  try
    for i = 0 to h.numelems do
      if not (should_move h (l_son i) i) then
	failwith "Bad heap!";
      if not (should_move h (r_son i) i) then
	failwith "Bad heap!"
    done
  with Incoherent_heap -> ();
*)
	
let iter f h = 
  for i=0 to h.numelems-1 do
    f (at_index h i)
  done;

