(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

{
open Parser;;        (* The type token is defined in parser.mli *)
open Error;;
open Big_int;;
open Pp;;
exception Eof;;
exception Illegal;;

let newline ()= 
  incr linenum;
  linestart:=Lexing.lexeme_end (!Error.mylexbuf);
;;
let symboltable = Hashtbl.create 100;;
let invsymboltable = Hashtbl.create 100;;
let lastsymbol = ref (-1);;
let int_of_symbol s = 
  try Hashtbl.find symboltable s 
  with Not_found -> 
    incr lastsymbol;
    Hashtbl.add symboltable s !lastsymbol;
    Hashtbl.add invsymboltable !lastsymbol s;
    !lastsymbol;;
let symbol_of_int = Hashtbl.find invsymboltable;;

let countlines s =
  for i=0 to (String.length s) - 1 do
    if s.[i] = '\n' then (newline ())
  done;;

let base2_16 a base=
  let b=ref (Big_int.big_int_of_int 0) in
  let c=ref 0 in
  for i=0 to (String.length a) - 1 do
    b:=Big_int.mult_int_big_int base !b;
    c:=int_of_char (a.[i]);
    if (!c>=48)&&(!c<=57) then c:=!c-48;
    if (!c>=65)&&(!c<=70) then c:=!c-55;
    if (!c>=97)&&(!c<=102) then c:=!c-87;
    b:=Big_int.add_int_big_int !c !b;
  done;
  !b;;

let getpos ()={ line= !linenum; col=(Lexing.lexeme_start !Error.mylexbuf) - (!linestart) };;
Hashtbl.add symboltable "temp" (-2);;
Hashtbl.add invsymboltable (-2) "temp";;

let string_after = fun s i -> String.sub s i ((String.length s) - i);;
let string_before = fun s i -> String.sub s 0 i;;

}
rule token = parse
    [' ' '\t' '\r' ] 
                   { token lexbuf }     (* skip blanks *) 
  | '\n'           { newline (); token lexbuf }
  | "do"           { DO (getpos ()) }
  | "if"           { IF (getpos ()) }
  | "else"         { ELSE (getpos ()) }
  | "while"        { WHILE (getpos ()) }
  | "do"           { DO (getpos()) }
  | "for"          { FOR (getpos()) }
  | "switch"       { SWITCH (getpos()) }
  | "case"         { CASE (getpos()) }
  | "void"         { VOID (getpos()) }
  | "byte"         { BYTE (getpos()) }
  | "continue"     { CONTINUE (getpos()) }
  | "break"        { BREAK (getpos()) }
  | "return"       { RETURN (getpos()) }
  | "static"       { STATIC (getpos()) }
  | "shr"          { SHR (getpos()) }
  | "shl"          { SHL (getpos()) }
  | "rol"          { ROL (getpos()) }
  | "ror"          { ROR (getpos()) }
  | "ro4"          { RO4 (getpos()) }
  | "!RA"          { TERA (getpos()) }
  | "!RB"          { TERB (getpos()) }
  | "!RC"          { TERC (getpos()) }
  | "!OPTION"      { TOPTION (getpos()) }
  | "uint"['1'-'9']['0'-'9']*
                   { UINTEGER ((getpos()),(int_of_string (string_after (Lexing.lexeme lexbuf)  4 ))) }
  | "sint"['1'-'9']['0'-'9']*
                   { SINTEGER ((getpos()),(int_of_string (string_after (Lexing.lexeme lexbuf)  4 ))) }
  | ("boolean" | "bit")
                   { BOOLEAN (getpos()) }
  | '"' ( [^ '"' ] | "\\\"" ) * '"'
                   { STRING ((getpos()),Lexing.lexeme lexbuf) }
  | ['A'-'Z''a'-'z'] ['A'-'Z''a'-'z''0'-'9'] * 
                   { IDENT ((getpos()),(int_of_symbol (Lexing.lexeme lexbuf))) }
  | ['0'-'9']+     { INT ((getpos()),(Big_int.big_int_of_string (Lexing.lexeme lexbuf))) }
  | "0x"['A'-'Z''a'-'z''0'-'9']+ { 
    INT((getpos()),base2_16 (string_after (Lexing.lexeme lexbuf) 2 ) 16) 
    }
  | (['0'-'9']['A'-'Z''a'-'z''0'-'9']*'h') {
    INT((getpos()),base2_16 (string_before (Lexing.lexeme lexbuf) ((String.length (Lexing.lexeme lexbuf))-1) ) 16) 
    }
  | "0b"['0'-'1']+ { 
    INT((getpos()),base2_16 (string_after (Lexing.lexeme lexbuf) 2 ) 2) 
    }
  | (['0'-'1']+'b') { 
    INT((getpos()),base2_16 (string_before (Lexing.lexeme lexbuf) ((String.length (Lexing.lexeme lexbuf))-1) ) 2) 
    }
  | "=="           { EQUALEQUAL (getpos()) }
  | "!="           { LESSGREATER (getpos()) }
  | "<="           { LESSEQUAL (getpos()) }
  | ">="           { GREATEREQUAL (getpos()) }
  | '<'            { LESS (getpos()) }
  | '>'            { GREATER (getpos()) }
  | ";"            { SEMI (getpos()) }
  | ","            { COMMA (getpos()) }
  | ':'            { COLON (getpos()) }
  | "::"           { COLONCOLON (getpos()) }
  | '?'            { QUESTIONMARK (getpos()) }
  | '.'            { DOT (getpos()) }
  | ".."           { DOTDOT (getpos()) }
  | '='            { EQUAL (getpos()) }
  | '-'            { MINUS (getpos()) }
  | '+'            { PLUS (getpos()) }
  | '*'            { TIMES (getpos()) }
  | '/'            { DIV (getpos()) }
  | '%'            { MOD (getpos()) }
  | "||"           { LOGICALOR (getpos()) }
  | '|'            { OR (getpos()) }
  | "&&"           { LOGICALAND (getpos()) }
  | '&'            { AND (getpos()) }
  | '^'            { XOR (getpos()) }  
  | '!'            { NOT (getpos()) }
  | '~'            { NEG (getpos()) }
  | "+="           { PLUSEQUAL (getpos()) }
  | "-="           { MINUSEQUAL (getpos()) }
  | "*="           { TIMESEQUAL (getpos()) }
  | "/="           { DIVEQUAL (getpos()) }
  | "&="           { ANDEQUAL (getpos()) }
  | "|="           { OREQUAL (getpos()) }
  | "^="           { XOREQUAL (getpos()) }
  | "%="           { MODEQUAL (getpos()) }
  | "++"           { INC (getpos()) }
  | "--"           { DEC (getpos()) }  
  | '('            { LPAREN (getpos()) }
  | ')'            { RPAREN (getpos()) }
  | '['            { LBRACKET (getpos()) }
  | ']'            { RBRACKET (getpos()) }
  | '{'            { LACCOLADE (getpos()) }
  | '}'            { RACCOLADE (getpos()) }
  | "/*"([^'*'] | "*"[^'/']) * "*/"
                   { countlines(Lexing.lexeme lexbuf); token lexbuf }   (* commentaires *)
  | "//" ([^ '\n' ])* 
                   { token lexbuf }
  | eof            { TEOF }
  | _              { parse_error("invalid characters : "^(Lexing.lexeme lexbuf)); token lexbuf }



