(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Smallset;;
open Gen;;
open Ass;;

(* Liveness analysis *)

let dead_oper_string instr =
  let indentlen = String.length Ass.indent in
  let rec ins_dead s n = 
    try 
      let i = (String.index_from s n '\n') in 
      (String.sub s n (1 + i +indentlen - n))^"; DEAD "^(ins_dead s (i + 1))
    with Not_found -> String.sub s n ((String.length s) - n)
  in
  "; DEAD "^(ins_dead (Ass.to_string Codegen.namer instr) 0);;

let flowgraph code =
  let labels = Hashtbl.create 101 in
  let addlabels = List.iter (function
      {instr = Ass.Label (_, l)} as n -> Hashtbl.add labels l n
    | _ -> ()) 
  in let node_of_instr i = 
    let use = Codegen.usedby i in
    let def = Codegen.defby i in
    { instr = i; use = of_list use; def = of_list def; 
      live_in = empty; live_out = empty; succ = []; dead = false }
  in let graph = List.map node_of_instr code in
  let rec fill_succ = function
      instr::succ::t -> instr.succ <- of_list [succ]; fill_succ (succ::t)
    | h::t -> fill_succ t
    | _ -> ()
  in let rec fill_jump_succ = function
      { instr = Ass.Oper (_, _, _, n, l) } as instr::t -> 
	let lnodes = List.map (function l -> Hashtbl.find labels l) l in
	let fnode = if n then instr.succ else empty in
	instr.succ <- union fnode (of_list lnodes); fill_jump_succ t
    | h::t -> fill_jump_succ t
    | _ -> ()
  in
  addlabels graph;
  fill_succ graph;
  fill_jump_succ graph;
  graph;;

let print_function_flow f =
  let print_intsp i = print_string (Codegen.namer i); print_char ' ' in
  print_newline ();
  List.iter (function c -> 
    if c.dead then print_string (Ass.indent^(dead_oper_string c.instr))
    else Ass.print stdout Codegen.namer c.instr; 
    print_string " ; Use ";
    List.iter print_intsp c.use;
    print_string " ; Def ";
    List.iter print_intsp c.def;
    print_string " ; In ";
    List.iter print_intsp c.live_in;
    print_string " ; Out ";
    List.iter print_intsp c.live_out;
    print_newline ()) f
;;

let print_program_flow p = 
  List.iter (function { Ass.flow = c } -> 
    print_function_flow c) p;;

(* let print_instr = function
    Ass.Oper (s, _, _, _) -> print_string s
  | Ass.Label (s, _) -> print_string s
  | Ass.Move (s, _, _) -> print_string s

let rec print_graph = 
    let print_intsp i = print_int i; print_char ' ' in
    function
	h::t -> 
	  print_instr h.instr; 
	  print_string " ; use ";
	  List.map print_intsp h.use;
	  print_string " ; def ";
	  List.map print_intsp h.def;
	  print_string " ; in ";
	  List.map print_intsp h.live_in;
	  print_string " ; out ";
	  List.map print_intsp h.live_out;
	  print_newline (); 
	  print_graph t
      | [] -> ()*)

let fixpoint graph =
  let finished = ref false in
  let reversed = List.rev graph in
  let update n = 
    let live_out = union_list ((List.map (function n -> n.live_in)) n.succ) in
    let live_in = union n.use (diff live_out n.def) in
    if !finished && (live_out <> n.live_out) then finished := false;
    n.live_out <- live_out;
    n.live_in <- live_in
  in
  while not !finished do
    finished := true;
(*    print_graph graph;
    print_newline ();*)
    List.iter update reversed
  done;
(*  print_graph graph;*)
  graph;;

let flow c = fixpoint (flowgraph c);;

let killer_fixpoint always_used graph =
  let always_used = of_list always_used in
  let finished = ref false in
  let reversed = List.rev graph in
  let kill n = n.dead <- true; finished := false in
  let killable n = not n.dead && match n.instr with
    Move _ -> true
  | Oper (_, _, _, true, []) -> true
  | _ -> false
  in
  let update n = 
    if (killable n) && ((inter n.live_out n.def) = []) && ((inter always_used n.def) = []) then
      kill n;
    let live_out = union_list ((List.map (function n -> n.live_in)) n.succ) in
    let live_in = if n.dead then live_out else union n.use (diff live_out n.def) in
    if !finished && (live_out <> n.live_out) then finished := false;
    n.live_out <- live_out;
    n.live_in <- live_in
  in
  while not !finished do
    if Args.verbose > 1 then begin
      print_string "\nKiller iteration *************************************************\n";
      print_function_flow graph;
    end;
    finished := true;
    List.iter update reversed
  done;
  graph;;

let dead_code_killer always_used flow = killer_fixpoint always_used flow;;

let code_of_graph graph = 
  List.map 
    (fun n -> if n.dead then 
      Oper (dead_oper_string n.instr, [], [], true, []) else n.instr)
    graph;;

(*let strip_succ = List.map (function n -> n.succ <- []; n);;
strip_succ (flow [17; 23] b);;
flow b;;
print_graph (flowgraph b);;
fixpoint (flowgraph b);;
strip_succ (fixpoint (flowgraph b));;*)

(* Dead code elimination *)

