(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

(* calcul de la rpartition en pages*)
open Tracetype;;
open Format;;

let mnblock=ref 0;

type pagebloc={
  bcallsize : int;
  bmincallsize : int;
  bjmpsize : int;
  items : pageitem list;
  mutable pnodes : node list;
};;

let emptypagebloc={bcallsize=0;bmincallsize=0;bjmpsize=0;items=[];pnodes=[]};;

let create_tree prg chan = (* prg de type tracetype.program*)
  let pagetable=Hashtbl.create 101 in 
  let blocks=Array.create pagecount emptypagebloc in
  let compare_link2 l1 l2 = (l1.weight2 < l2.weight2) in
  let index_link2 l1 = (l1.node1.pid,l1.node2.pid) in

  let processnodelist = ref [] in
  let pageitem_id= ref (-1) in
  let linkheap = ref (Heap.create (compare_link2) (index_link2) 101) in
  let mainlabel = Gen.label_of_name "_main" in

  let display_lnk lnk2=(
    printf "%d-%d:%3.1f " lnk2.node1.pid lnk2.node2.pid lnk2.weight2
  ) in

  let display_heap () =(
    for i=0 to (!linkheap).Heap.numelems-1 do
      match (Array.get (!linkheap).Heap.heap i) with
	| None -> ();
	| Some lnk -> display_lnk lnk;
    done;
    printf "\n";
  ) in

  let display_pageitem node=(
    printf "node %d ->(%s %s) : jmp %4d  call %4d  callmin %4d  %3.1f %s\n" node.pid 
			       (match node.child1 with None -> "none" | Some a -> sprintf "%4d" a.pid ) 
			       (match node.child2 with None -> "none" | Some a -> sprintf "%4d" a.pid ) 
			       node.jmpsize node.callsize node.callminsize node.internalconstraint
                               (if node.contain_main then "<main" else "");
    printf "  nodes   =[%s]\n" (Ast.slist ", " (function a->sprintf "%d" a.label.Gen.id) node.nodelist);  
    printf "  extlinks=[%s]\n" (Ast.slist ", " (function lnk2->sprintf "%d-%d:%3.1f" lnk2.node1.pid lnk2.node2.pid lnk2.weight2) node.extlinklist);    
    printf "  arglinks=[%s]\n" (Ast.slist ", " (function wn2->sprintf "%d:%3.1f" wn2.ssnode.pid wn2.ssweight) node.arglinklist);    
  ) in

  let addlink2 lnk2 = ( (*cette fonction suppose lnk2.node1.pid>lnk2.node2.pid*)
    if (lnk2.node1.pid<=lnk2.node2.pid) then printf "invalid pids in addlink2 %d<= %d\n" lnk2.node1.pid lnk2.node2.pid else
      let robj=Heap.collision (!linkheap) lnk2 in
	begin
	  (* insertion d'un lment. Si dj dans la liste, on ajoute les poids *)
	  match robj with
	    | None -> (
		Heap.add (!linkheap) lnk2;
		lnk2.node1.extlinklist<-lnk2::lnk2.node1.extlinklist;
		lnk2.node2.extlinklist<-lnk2::lnk2.node2.extlinklist;  
		if (Args.verbose>=2) then printf "new link %d-%d weight=%3.1f\n" lnk2.node1.pid lnk2.node2.pid lnk2.weight2;
	      )
	    | Some obj -> (
		Heap.remove (!linkheap) obj;
		(*fusionne lnk2 et obj *)
		if (Args.verbose>=2) then printf "update link %d-%d weight=%3.1f+%3.1f=%3.1f\n" lnk2.node1.pid lnk2.node2.pid obj.weight2 lnk2.weight2 (lnk2.weight2+.obj.weight2);
		obj.weight2 <- lnk2.weight2+.obj.weight2;
		Heap.add (!linkheap) obj; (*ajoute l'lment fusionn dans la liste*)
	      )
	end 
  ) in

  let addlink lnk = (
    if (Args.verbose>=4) then display_heap();
    if (lnk.src<>lnk.dest) then 
    let comp=(lnk.src.page_item.pid-lnk.dest.page_item.pid) in 
    let nd1=if (comp>0) then lnk.src else lnk.dest in  (* convention: le premier noeud est toujours de pid le plus important (pour avoir un graphe non orient) *)
    let nd2=if (comp>0) then lnk.dest else lnk.src in
    let lnk2={node1=nd1.page_item;node2=nd2.page_item;weight2=lnk.weight} in
      addlink2 lnk2;
  ) in

  let addlinks linklist = (
    List.iter (function lnk -> addlink lnk ) linklist;
  ) in
    
  let node_to_pageitem anode=(
    incr pageitem_id;
    if (Args.verbose>=3) then printf "add node %d of label %s\n" (!pageitem_id) anode.label.Gen.name;
    let mypageitem={ 
      pid=(!pageitem_id);
      nodelist = [anode];
      arglinklist = [];  
      extlinklist = [];  (* mis  jour par la passe sur les links*)
      child1 = None;
      child2 = None;
      jmpsize = (if anode.callable then 0 else anode.maxsize);
      callsize = (if anode.callable then anode.maxsize else 0);
      callminsize = (if anode.callable then instr_max_words (Call {src=anode;dest=anode;weight=0.}) else 0);
      contain_main = (anode.label=mainlabel);
      internalconstraint=0.;
    } in (
	anode.page_item<-mypageitem;  
	if (Args.verbose>=2) then display_pageitem mypageitem;
	mypageitem;
      )
  ) in
    
  let min i1 i2 = (if (i1<i2) then i1 else i2 ) in

  let updatelink init final lnk suppr= (  (* change le lien init-other en final-other *) (*suppose que final.id > init.id*)
    (try (Heap.remove (!linkheap) lnk);(if Args.verbose>=2 then printf "link %d-%d removed\n" lnk.node1.pid lnk.node2.pid) with
      Not_found->());
    (*supprimer le lien lnk de la liste extlink de other  *)
    if (lnk.node1=init) then (
      lnk.node2.extlinklist<-List.filter (function l->lnk<>l) lnk.node2.extlinklist; 
      if (lnk.node2<>suppr) then addlink2 {node1=final;node2=lnk.node2;weight2=lnk.weight2}; 
    ) else (      
      lnk.node1.extlinklist<-List.filter (function l->lnk<>l) lnk.node1.extlinklist;
      if (lnk.node1<>suppr) then addlink2 {node1=final;node2=lnk.node1;weight2=lnk.weight2}; 
    )
  ) in

  let rec union_sorted l1 l2 = ( match l1, l2 with  (*suppose listes tries par pid croissant*)
      h1::t1, (h2::t2) when h1.ssnode.pid=h2.ssnode.pid   -> union_sorted t1 ({ssnode=h2.ssnode;ssweight=h1.ssweight+.h2.ssweight}::t2)
    | h1::_ as l, h2::t when h1.ssnode.pid > h2.ssnode.pid -> h2::(union_sorted l t)
    | h1::t, (h2::_ as l) -> h1::(union_sorted t l)
    | [], l -> l
    | l, [] -> l ) 
  in
  let union_sort l1 l2 = (
    union_sorted (List.sort (fun ln1 ln2-> ln1.ssnode.pid-ln2.ssnode.pid) l1) (List.sort (fun ln1 ln2-> ln1.ssnode.pid-ln2.ssnode.pid) l2)) in
  let rec isect_sorted l1 l2 = ( match l1, l2 with  (*suppose listes tries par pid croissant*)
      h1::t1, (h2::t2) when h1.pid=h2.ssnode.pid   -> h2::(isect_sorted t1 t2)
    | h1::_ as l, h2::t when h1.pid > h2.ssnode.pid -> isect_sorted l t
    | h1::t, (h2::_ as l) -> isect_sorted t l
    | [], l -> []
    | l, [] -> [] ) 
  in
  let isect_sort l1 l2 = 
    isect_sorted (List.sort (fun ln1 ln2-> ln1.pid-ln2.pid) l1) (List.sort (fun ln1 ln2-> ln1.ssnode.pid-ln2.ssnode.pid) l2); 

  in let discarded=ref []

  in let findfreeblock myblocks myroot =(
      let utility blc aroot = (
        List.fold_left (fun nb wnode->nb+.wnode.ssweight) 0. (isect_sort blc.items aroot.arglinklist);
      ) in
      let found=ref (-1) in 
      let utilmax=ref (-1.) in
	(
          for t=0 to pagecount-1 do (*recherche du meilleur bloc libre*)
	    if ((myblocks.(t).bjmpsize+myroot.jmpsize+myblocks.(t).bcallsize+myroot.callsize)<=pagemax)&&
	      ((min (myblocks.(t).bcallsize+myroot.callminsize) (myroot.callsize+myblocks.(t).bmincallsize))<pagecallmax)&&
	      ((utility myblocks.(t) myroot)>(!utilmax)) then (
		  utilmax:=(utility myblocks.(t) myroot);
		  found:=t;
		)
	  done;
	  !found;
	)
    )
  in let addtoblock myblocks id myroot =
      let blc=myblocks.(id) in
	myblocks.(id)<-{
	  bjmpsize=blc.bjmpsize+myroot.jmpsize;
	  bcallsize=blc.bcallsize+myroot.callsize;
	  bmincallsize=min (blc.bcallsize+myroot.callminsize) (myroot.callsize+blc.bmincallsize);
	  items=myroot::blc.items;
	  pnodes=[];
	}
    					 
  in let rec place_in_blocks myblocks myroot=(
      let id=findfreeblock myblocks myroot in (
	  if (id>=0) then addtoblock myblocks id myroot 
	  else (
            match (myroot.child1,myroot.child2) with
	      | (Some child1,Some child2) ->  place_in_blocks myblocks child1;place_in_blocks myblocks child2;
	      | (_,_) -> discarded:=myroot::(!discarded);
	  )
	)
    ) 
  in let lastlabel = ref "" 
  in let lastinstr = ref Ret (*cette valeur force l'affichage du premier label *)
  in let output_code chan =(			       
      Printf.fprintf chan "    reset _main\n\n";
      Printf.fprintf chan "    device  PINS28,OSCXT2,turbo,stackx,optionx,carryx\n";
      Printf.fprintf chan "WW    equ 01h\n";
      for r=0 to 255 do
	if (r mod 32) >= 16 || (r >= 8 && r < 16) then
	  Printf.fprintf chan "r0x%02X equ %02Xh\n" r (r mod 32);
      done;
      for t=0 to (pagecount-1) do
        lastinstr:=Ret;
	Printf.fprintf chan "\n;**********************************************************************\n";
	Printf.fprintf chan "    Org 0%Xh          ; page %d\n\n" (t*pageaddress) t;
	List.iter (
	  function n -> (
	    (match !lastinstr with 
               | Ret -> ();
	       | Jmp lnk -> if not (lnk.dest.label=n.label) then Printf.fprintf chan "%s\n" (jinstr_string t (function lab->Hashtbl.find pagetable lab) !lastinstr ) 
	       | Jcond (str,lnk1,lnk2) -> if (lnk1.dest.label=n.label) then (
		   Printf.fprintf chan "%s\n" (jinstr_stringbool false  t (function lab->Hashtbl.find pagetable lab) !lastinstr);
		 ) else if (lnk2.dest.label=n.label) then (
		   Printf.fprintf chan "%s\n" (jinstr_stringbool true  t (function lab->Hashtbl.find pagetable lab) !lastinstr);
		 ) else (
		   Printf.fprintf chan "%s\n" (jinstr_string  t (function lab->Hashtbl.find pagetable lab) !lastinstr);
		 ) 

	    );
	    Printf.fprintf chan "%s\n" (Gen.label_string n.label);
	    List.iter (fun i -> Printf.fprintf chan "%s\n" (instr_string t (function lab->Hashtbl.find pagetable lab) i)) n.code;
	    lastinstr:=n.lastinstr; (* pour supprimer optimisation commenter *)
	    (match !lastinstr with 
	       | Ret -> Printf.fprintf chan "%s\n" (jinstr_string  t (function lab->Hashtbl.find pagetable lab) !lastinstr);
	       | _ -> ());
	  )
	) blocks.(t).pnodes;
	match !lastinstr with 
	   | Ret -> ();
	   | _ -> Printf.fprintf chan "%s\n" (jinstr_string  t (function lab->Hashtbl.find pagetable lab) !lastinstr);
      done;  
    )
  in (
    if (Args.verbose>=1) then (
      printf "Start of page repartition ***************************************************\n";
      printf "\n Building initial structure *******\n";
    );
    (*cre la liste initiale de noeuds *)
    processnodelist:=List.fold_left (fun lst item -> node_to_pageitem item::lst ) [] prg.nodes; 
    (*cre la liste initiale de links: chaque link est ajout en double*)
    List.iter (function anode -> addlinks anode.links) prg.nodes;
    (*met  jour les poids entre les diffrents sous-arbre (de taille 1) *)
    Heap.iter (function lnk2-> 
      lnk2.node1.arglinklist<-{ssnode=lnk2.node2;ssweight=lnk2.weight2}::lnk2.node1.arglinklist;
      lnk2.node2.arglinklist<-{ssnode=lnk2.node1;ssweight=lnk2.weight2}::lnk2.node2.arglinklist;
    ) (!linkheap);
    List.iter (function anode -> anode.arglinklist<-(List.sort (fun ln1 ln2-> ln1.ssnode.pid-ln2.ssnode.pid) anode.arglinklist)) (!processnodelist);
    if (Args.verbose>=1) then (
      printf "\n Initial structure *****************\n";
      List.iter (function anode->display_pageitem anode) (!processnodelist);
      printf "\n Merging ***************************\n";
    );
    (*processus d'aggrgation*)  
    let count=ref 0 in
    while (not (Heap.empty (!linkheap))) do 
      begin
        incr count;
        if (Args.verbose>=1) then printf "\nStep %d\n" (!count);
	if (Args.verbose>=2) then display_heap ();
	incr pageitem_id;
	(*prend le lien le plus fort et aggrge les noeuds correspondants *)
	let hlink=Heap.head (!linkheap) in
	let node1=if (hlink.node1.internalconstraint>hlink.node2.internalconstraint) then hlink.node1 else hlink.node2 in (*node1 le plus li*)
	let node2=if (hlink.node1.internalconstraint>hlink.node2.internalconstraint) then hlink.node2 else hlink.node1 in (*node2 le moins li*)
	let node={ 
          pid= (!pageitem_id);
	  nodelist= node1.nodelist@node2.nodelist;
	  arglinklist= union_sort node1.arglinklist node2.arglinklist;
          extlinklist=[]; (* mis  jour plus loin *)
	  child1=Some node1;
	  child2=Some node2;
	  
	  jmpsize=node1.jmpsize+node2.jmpsize;
	  callminsize=min (node1.callsize+node2.callminsize) (node2.callsize+node1.callminsize);
	  callsize=node1.callsize+node2.callsize;
	  contain_main=node1.contain_main||node2.contain_main;
	  internalconstraint=hlink.weight2;
	} in begin 
            if (Args.verbose>=1) then printf "fusion %d with %d -> %d\n" node1.pid node2.pid node.pid;
            (* change les liens de type node1-* et node2-* en node-* *)
	    List.iter (function lnk->updatelink node1 node lnk node2) node1.extlinklist;
	    List.iter (function lnk->updatelink node2 node lnk node1) node2.extlinklist;
            node1.extlinklist<-[];
            node2.extlinklist<-[];
            (* met  jour arglink list des autres lments *)
            List.iter (function wnode-> wnode.ssnode.arglinklist<-{ssnode=node;ssweight=wnode.ssweight}::wnode.ssnode.arglinklist  ) node.arglinklist;  
            if (Args.verbose>=2) then ( display_pageitem node;
					display_pageitem node1; 
 					display_pageitem node2;);
            processnodelist:=node::(!processnodelist);
 	  end


      end;
    done;
    if (Args.verbose>=1) then (
      printf "\n Merge results ***************************\n";
      List.iter (function anode->display_pageitem anode) (!processnodelist);
      printf "\n Allocating    ***************************\n";
    );
    for t=0 to (pagecount-1) do blocks.(t)<-{ bcallsize =0;bmincallsize=0;bjmpsize=(if t=(pagecount-1) then 1 else 0(*pour le jmp de dbut*));items=[];pnodes=[]} done;
    place_in_blocks blocks (List.hd (!processnodelist));
    if (!discarded<>[]) then (Printf.fprintf stderr "Error : can't fit the program in memory \n";exit 1;);
    (* cherche la page de _main si autre que 0*)
    mnblock:=(-1);
    for t=0 to (pagecount-1) do 
      blocks.(t).pnodes<-List.fold_left (fun lst a->(List.filter (function anode->not anode.callable) a.nodelist)@lst) [] blocks.(t).items; 
      blocks.(t).pnodes<-List.fold_left (fun lst a->(List.filter (function anode->anode.callable) a.nodelist)@lst) blocks.(t).pnodes blocks.(t).items;
      List.iter (function anode->if (anode.contain_main) then mnblock:=t;) blocks.(t).items;
    done;
    if (!mnblock<0) then printf "Error : can't find _main" else
      let temp=blocks.(!mnblock) in
	if (!mnblock>0) then (
	  (*swap blocks*)
	  blocks.(!mnblock)<-blocks.(0);
	  blocks.(0)<-temp;
	);
    for t=0 to (pagecount-1) do 
      List.iter (function anode->Hashtbl.add pagetable anode.label t) blocks.(t).pnodes;
    done;
    if (Args.verbose>=1) then 	(
      for t=0 to (pagecount-1) do
	printf "  block %d : jmp %4d  call %4d  callmin %4d \n" t blocks.(t).bjmpsize blocks.(t).bcallsize blocks.(t).bmincallsize;
	printf "    %s \n" (Ast.slist ", " (function a->sprintf "n%d" a.pid) blocks.(t).items);  
	printf "    %s \n\n" (Ast.slist ", " (function a->(Ast.slist ", " (function a->sprintf "%d" a.label.Gen.id) a.nodelist)) blocks.(t).items);  
      done;
      printf "\n******************************************************************************************\n";
      printf "****************************** Final code ************************************************\n";
      printf "******************************************************************************************\n";
    );
    output_code chan;
)
    (*resultat dans processnodelist *)
;;

