(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Tracetype;;

let gengraph fasm = (* fills label, code, lastinstr, callable,  *)
  let label_nodes = Hashtbl.create 101 in
  let node_of_label = Hashtbl.find label_nodes in

  let gentraces asm =
    let addtrace label tr last =
      let n = new_node label tr last
      in
      Hashtbl.add label_nodes label n
    in
    let rasm = List.rev asm in
    let rec gentraces tr last = function
	[] -> ()
      |	instr::rasm -> 
	  match instr, last with
	    Fasm.Label l, Some last -> 
	      addtrace l tr last;
	      gentraces [] None rasm
	  | Fasm.Label l, none ->
	      failwith ("Unterminated block starting at "^l.Gen.name^" in Traceprep.gengraph")
	  | Fasm.Ret, _ | Fasm.Jmp _, _ | Fasm.Jcond _, _ -> gentraces [] (Some instr) rasm
	  | Fasm.Oper (_, 0), _ -> gentraces tr last rasm
	  | _ -> gentraces (instr::tr) last rasm
    in
    gentraces [] None rasm
  in

  let connectgraph () =
    let nodes = ref [] in
    let tofill = Queue.create () in

    let addlink f l w =
      let n = node_of_label l in
      if not n.flag then (n.flag <- true; Queue.add n tofill);
      let lnk = { src = f; dest = node_of_label l; weight = w } in
      n.links <- lnk::n.links;
      n.links <- lnk::n.links;
      lnk 
    in    
    let fillnode n =
      let transcode_instr = function
	  Fasm.Oper (s, i) -> Oper (s, i)
	| Fasm.Call l -> (node_of_label l).callable <- true; Call (addlink n l 1.)
	| Fasm.Bank i -> Bank i
	| _ -> failwith "Called Traceprep...transcode_instr with bad data"
      in
      let transcode_jinstr = function
	  Fasm.Ret -> Ret
	| Fasm.Jmp l -> Jmp (addlink n l 1.)
	| Fasm.Jcond (c, l1, l2) -> Jcond (c, (addlink n l1 1.), (addlink n l2 1.))
	| _ -> failwith "Called Traceprep...transcode_jinstr with bad data"
      in
      nodes := n::!nodes;
      n.code <- List.map transcode_instr n.precode;
      n.lastinstr <- transcode_jinstr n.prelastinstr;
    in

    let main = node_of_label (Gen.label_of_name "_main") in
    main.flag <- true;
    fillnode main;
    (try while true do
      let n = Queue.take tofill in
      fillnode n;
    done with Queue.Empty -> ());
    node_of_label, List.rev !nodes (* Rev so that predecessors tend to come first *)
  in

  let findfuncs nodes =
    let funcs = ref [] in
    let add f n = 
      if List.mem n f.callers then () else f.callers <- n::f.callers;
      if List.mem f !funcs then () else funcs := f::!funcs 
    in
    List.iter 
      (fun n -> 
	ignore (List.fold_left (function callnum -> function
	    Call { dest = d } -> add d (n, callnum); callnum + 1
	  | _-> callnum) 0 n.code)) 
      nodes;
    !funcs;
  in

  let add_func_info nodes funcs =
    let setfunc (f : node) (n : node) = 
      match n.belongfunc with
	None -> n.belongfunc <- Some f; true
      | Some n when n = f -> false
      | _ -> failwith "Flow node is in many functions!"
    in
    let update n = 
      List.filter 
	(match n.belongfunc with
	  Some f -> setfunc f
	| None -> function x -> false)
	(successors n)
    in
    List.iter (fun n -> ignore (setfunc n n)) funcs;
    fixpoint update nodes
  in

  gentraces fasm;
  let node_of_label, nodes = connectgraph () in
  let funcs = findfuncs nodes in
  add_func_info nodes funcs;
  { node_of_label = node_of_label; nodes = nodes; funcs = funcs }
;;

let fill_maxsize { nodes = nodes } =
  List.iter 
    (function n ->
      n.maxsize <- (jinstr_max_words n.lastinstr) +
	(List.fold_left (fun n i -> n + (instr_max_words i)) 0 n.code)
      ) nodes
;;

let output_node chan n =
  Printf.fprintf chan "%s ; %s\n" (Gen.label_string n.label)
    (match n.belongfunc with 
      None -> "not in a called function"
    | Some n -> 
	"in function : "^(Gen.label_string n.label));
  if n.callers != [] then begin
    output_string chan "; Caller nodes : ";
    List.iter (fun n -> Printf.fprintf chan "%s.%i " 
	(Gen.label_string (fst n).label) (snd n)) n.callers;
    output_char chan '\n';
  end;
  List.iter (fun i -> Printf.fprintf chan "%s\n" (instr_string 0 (fun _ -> 0) i)) n.code;
  Printf.fprintf chan "%s\n" (jinstr_string 0 (fun _ -> 0) n.lastinstr);;

let output_graph chan { nodes = nodes } =
  List.iter (output_node chan) nodes;;

let triv_bank_simplify { nodes = nodes } =
  let triv_bank_simplify n = 
    let rec simplify_instr rest curbank = match rest with
      [] -> []
    | inst::rest -> match curbank, inst with
	Some b, Bank b' when b = b' -> 
	  if (Args.verbose > 1) then
	  Printf.printf "Simplified a bank instruction.\n";
	  simplify_instr rest curbank
      | _, Bank b -> inst::(simplify_instr rest (Some b))
      | _, Call _ -> inst::(simplify_instr rest None)
      | _ -> inst::(simplify_instr rest curbank)
    in
    n.code <- simplify_instr n.code None
  in
  List.iter triv_bank_simplify nodes
;;

(*  Old snippets


 let trychange f = List.fold_left (fun changed n -> f n || changed) false in
      
  let fixpoint update l =
    while trychange update l do () done
  in

  let add_func_info nodes funcs =
    let setfunc (f : node) (n : node) = 
      match n.belongfunc with
	None -> n.belongfunc <- Some f; true
      | Some n when n = f -> false
      | _ -> failwith "Flow node is in many functions!"
    in
    let update n = match n.belongfunc with
      Some f -> trychange (setfunc f) (successors n)
    | None -> false
    in
    List.iter (fun n -> ignore (setfunc n n)) funcs;
    fixpoint update nodes
  in*)

