(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Format;;

let pagecount=4;;
let pagecallmax=256;;
let pagemax=512;;
let pageaddress=512;;

type 'a eoption = ENone | EMany | ESome of 'a

type instr = 
    Oper of string * int
  | Call of link
  | Bank of int

and jinstr = 
  | Jcond of string * link * link
  | Jmp of link
  | Ret

and link = {
    src : node;
    dest : node;
    mutable weight : float;
  } 

and node = {
    label : Gen.label; (* OK *)

(* code *) 
    mutable code : instr list; (* OK *)
    mutable lastinstr : jinstr; (* OK *)

(* graph information *)
    mutable callable : bool; (* OK *)
(*    mutable pred : node list; *)
    mutable links : link list; (* All links to or from node *) (* OK *)
    mutable belongfunc : node option;
    mutable callers : (node * int) list;

(* code size information *)
    mutable maxsize : int; (* All included, being pessimistic *) (* used *)

(* graph generation *)
    precode : Fasm.instr list; (* OK *)
    prelastinstr : Fasm.instr; (* OK *)
    mutable flag : bool; (* OK *)

(* partitioning *)
    mutable page_item : pageitem; (* OK *)
  }

and pageitem = {
  nodelist : node list;
  mutable arglinklist : weightednode list; (* liens vers des sous graphes*)
  mutable extlinklist : link2 list; (* liens externes du noeud en question*)
  child1 : pageitem option; (* noeud le plus li *)
  child2 : pageitem option;
  jmpsize : int;
  callsize : int;
  callminsize : int;
  contain_main : bool;
  internalconstraint : float;
  pid :int;
} 
and link2 = {
  node1 : pageitem;
  node2 : pageitem;
  mutable weight2: float;
} 
and weightednode={
  ssnode : pageitem;
  ssweight : float; 
}

let null_pageitem = { 
  nodelist = []; arglinklist = []; extlinklist = []; child1 = None; child2 = None; 
  jmpsize = 0; callsize = 0; callminsize = 0; contain_main = false ;
  internalconstraint = 0.; pid = 0; }

let new_node label tr last = { 
  label = label; precode = tr; prelastinstr = last; flag = false; maxsize = -1;
  code = []; lastinstr = Ret; callable = false; links = []; page_item = null_pageitem;
  belongfunc = None; callers = []; }

type program = {
    node_of_label : Gen.label -> node;
      nodes : node list;
      funcs : node list;
  } 

let successors n =
  match n.lastinstr with
    Ret -> []
  | Jmp { dest = n} -> [n]
  | Jcond (_, { dest = n1 }, { dest = n2 }) -> [n1; n2]
;;

let instr_min_words = function
    Oper (_, i) -> i
  | Call l -> 1
  | Bank i -> 1
;;

let instr_max_words = function
    Oper (_, i) -> i
  | Call l -> 2
  | Bank i -> 1
;;

let jinstr_min_words = function
    Jmp _ -> 1
  | Jcond _ -> 2
  | Ret -> 1
;;

let jinstr_max_words = function
    Jmp _ -> 2
  | Jcond _ -> 5
  | Ret -> 1
;;

let compare_node n1 n2 = 
  compare n1.label.Gen.id n2.label.Gen.id

let instr_string pg lbl2pg = function
    Oper (s, _) -> s
  | Call n -> 
      let pg1=(lbl2pg n.dest.label) in
      let s=if (pg1<>pg) then sprintf "    page 0%Xh\n" (pg1*pageaddress) else "" in
	s^"    call "^(Gen.label_string n.dest.label)
  | Bank i -> "    bank #"^(string_of_int i)
;;

let jinstr_string pg lbl2pg instr= 
  match instr with
    Jmp n -> (
      let pg1=(lbl2pg n.dest.label) in
      let s=if (pg1<>pg) then sprintf "    page 0%Xh\n" (pg1*pageaddress) else "" in
      s^"    jmp "^(Gen.label_string n.dest.label);
    );
  | Jcond (s, n1, n2) -> 
      let pg1=(lbl2pg n1.dest.label) in
      let pg2=(lbl2pg n2.dest.label) in	
      let s1=if (pg1<>pg) then sprintf "    page 0%Xh\n" (pg1*pageaddress) else "" in
      let s2=if (pg2<>pg1) then sprintf "    page 0%Xh\n" (pg2*pageaddress) else "" in (	
	  s1^"    snb "^s^" ; as one jcond instruction "^
	  "\n    jmp "^(Gen.label_string n1.dest.label)^"\n"^s2^
	  "    jmp "^(Gen.label_string n2.dest.label)
	);
  | Ret -> "    retp"
;;

let jinstr_stringbool b pg lbl2pg= function
    Jmp n -> 
      let pg1=(lbl2pg n.dest.label) in
      let s=if (pg1<>pg) then sprintf "    page 0%Xh\n" (pg1*pageaddress) else "" in
      s^"    jmp "^(Gen.label_string n.dest.label)
  | Jcond (s, n1, n2) -> if b then (
      let pg1=(lbl2pg n1.dest.label) in
      let pg2=(lbl2pg n2.dest.label) in	
      let s1=if (pg1<>pg) then sprintf "    page 0%Xh\n" (pg1*pageaddress) else "" in
      let s2=if (pg2<>pg1) then sprintf "    page 0%Xh\n" (pg2*pageaddress) else "" in (	
	  s1^"    snb "^s^" ; as one jcond instruction "^
	  "\n    jmp "^(Gen.label_string n1.dest.label)^"\n"^s2^
	  ";       jmp "^(Gen.label_string n2.dest.label)
	);
    ) else (
      let pg1=(lbl2pg n1.dest.label) in
      let pg2=(lbl2pg n2.dest.label) in	
      let s1=if (pg1<>pg2) then sprintf "    page 0%Xh\n" (pg1*pageaddress) else "" in
      let s2=if (pg2<>pg) then sprintf "    page 0%Xh\n" (pg2*pageaddress) else "" in (	
	  s2^"    sb "^s^" ; as one jcond instruction "^
	  "\n    jmp "^(Gen.label_string n2.dest.label)^"\n"^s1^
	  ";       jmp "^(Gen.label_string n1.dest.label)
	);
    )
  | Ret -> "    retp"
;;
   
let chflag { nodes = nodes } newval =
  List.iter (fun n -> n.flag <- newval) nodes

let fixpoint update l =
  let trychange f = List.fold_left (fun changed n -> f n || changed) false in
  let rec explore changed = function
      [] -> changed
    | h::t -> 
	let nl = update h in
	if nl = [] then explore changed t
	else explore true nl
  in
  while trychange (fun e -> explore false [e]) l do () done
    
